# -*- coding: utf-8 -*-
"""
This script converts Chinese-CLIP text and vision ONNX models to TensorRT format for GPU deployment.
"""

import os
import argparse

import tensorrt as trt
from tensorrt.tensorrt import Logger, Runtime
from cn_clip.clip.utils import _MODEL_INFO
from tensorrt_utils import TensorRTShape, build_engine

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-arch", 
        required=True, 
        choices=["ViT-B-16", "ViT-L-14", "ViT-L-14-336", "ViT-H-14", "RN50"],
        help="Specify the architecture (model scale) of Chinese-CLIP model to be converted."
    )    
    parser.add_argument(
        "--convert-text",
        action="store_true",
        help="Whether to convert the text encoder (text feature extractor) into TensorRT."
    )
    parser.add_argument(
        "--text-onnx-path", 
        default=None, 
        type=str,
        help="If --convert-text is True, specify the path of the input text ONNX model."
    )
    parser.add_argument(
        "--convert-vision",
        action="store_true",
        help="Whether to convert the vision encoder (vision feature extractor) into TensorRT."
    )
    parser.add_argument(
        "--vision-onnx-path", 
        default=None, 
        type=str,
        help="If --convert-vision is True, specify the path of the input vision ONNX model."
    )
    parser.add_argument('--batch-size', default=1, type=int, help='The batch size of the TensorRT model')
    parser.add_argument(
        "--context-length", type=int, default=52, help="The padded length of input text (include [CLS] & [SEP] tokens)."
    )
    parser.add_argument(
        "--save-tensorrt-path", 
        required=True,
        type=str,
        help="Path (prefix) of the output converted TensorRT Chinese-CLIP text or vision engines."
    )
    parser.add_argument('--fp16', action='store_true',
                        help='Use when convert onnx to FP16 TensorRT model.')
    parser.add_argument('--fp32', action='store_true',
                        help='Use when convert onnx to FP32 TensorRT model.')
    parser.add_argument('--fp16-banned-ops', type=str, default='',
                        help='The ops need to be banned when building FP16 TensorRT model.')
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()

    # Log params.
    print("Params:")
    for name in sorted(vars(args)):
        val = getattr(args, name)
        print(f"  {name}: {val}")

    trt_logger: Logger = trt.Logger(trt.Logger.INFO)
    runtime: Runtime = trt.Runtime(trt_logger)

    trt.init_libnvinfer_plugins(trt_logger, '')

    # ONNX -> TensorRT
    batch_size = args.batch_size
    if args.convert_text:
        seq_len = args.context_length
        text_input_shape = [TensorRTShape((batch_size, seq_len),
                                        (batch_size, seq_len),
                                        (batch_size, seq_len), 'text')]
        input_text_onnx_path = args.text_onnx_path
        assert os.path.exists(input_text_onnx_path), f"Error: The specified --text-onnx-path {input_text_onnx_path} not exists!"
    
    if args.convert_vision:
        image_size = _MODEL_INFO[args.model_arch]['input_resolution']
        vision_input_shape = [TensorRTShape((batch_size, 3, image_size, image_size),
                                            (batch_size, 3, image_size, image_size),
                                            (batch_size, 3, image_size, image_size), 'image')]
        input_vision_onnx_path = args.vision_onnx_path
        assert os.path.exists(input_vision_onnx_path), f"Error: The specified --vision-onnx-path {input_vision_onnx_path} not exists!"

    if args.fp32:
        print("---------------build TensorRT engine: FP32------------------")
        if args.convert_text:
            engine = build_engine(
                runtime=runtime,
                onnx_file_path=input_text_onnx_path,
                logger=trt_logger,
                input_shapes=text_input_shape,
                workspace_size=10000 * 1024 * 1024,
                fp16=False,
                int8=False,
            )
            text_fp32_trt_path = f"{args.save_tensorrt_path}.txt.fp32.trt"
            print(f"Saving the text FP32 TensorRT model at {text_fp32_trt_path} ...")
            with open(text_fp32_trt_path, 'wb') as f:
                f.write(bytearray(engine.serialize()))
        if args.convert_vision:
            engine = build_engine(
                runtime=runtime,
                onnx_file_path=input_vision_onnx_path,
                logger=trt_logger,
                input_shapes=vision_input_shape,
                workspace_size=10000 * 1024 * 1024,
                fp16=False,
                int8=False,
            )
            vision_fp32_trt_path = f"{args.save_tensorrt_path}.img.fp32.trt"
            print(f"Saving the vision FP32 TensorRT model at {vision_fp32_trt_path} ...")
            with open(vision_fp32_trt_path, 'wb') as f:
                f.write(bytearray(engine.serialize()))
    
    if args.fp16:
        print("-----------------build TensoRT engine: FP16----------------")
        if args.convert_text:
            engine = build_engine(
                runtime=runtime,
                onnx_file_path=input_text_onnx_path,
                logger=trt_logger,
                input_shapes=text_input_shape,
                workspace_size=10000 * 1024 * 1024,
                fp16=True,
                int8=False,
            )
            text_fp16_trt_path = f"{args.save_tensorrt_path}.txt.fp16.trt"
            print(f"Saving the text FP16 TensorRT model at {text_fp16_trt_path} ...")
            with open(text_fp16_trt_path, 'wb') as f:
                f.write(bytearray(engine.serialize()))
        if args.convert_vision:
            engine = build_engine(
                runtime=runtime,
                onnx_file_path=input_vision_onnx_path,
                logger=trt_logger,
                input_shapes=vision_input_shape,
                workspace_size=10000 * 1024 * 1024,
                fp16=True,
                int8=False,
            )
            vision_fp16_trt_path = f"{args.save_tensorrt_path}.img.fp16.trt"
            print(f"Saving the vision FP16 TensorRT model at {vision_fp16_trt_path} ...")
            with open(vision_fp16_trt_path, 'wb') as f:
                f.write(bytearray(engine.serialize()))

    print("Finished ONNX to TensorRT conversion...")
    if args.convert_text:
        if args.fp32:
            print(f">>> The text FP32 TensorRT model is saved at {text_fp32_trt_path}")
        if args.fp16:
            print(f">>> The text FP16 TensorRT model is saved at {text_fp16_trt_path}")
    if args.convert_vision:
        if args.fp32:
            print(f">>> The vision FP32 TensorRT model is saved at {vision_fp32_trt_path}")
        if args.fp16:
            print(f">>> The vision FP16 TensorRT model is saved at {vision_fp16_trt_path}")


# Code modified from https://github.com/openai/CLIP

import json
import os
from pathlib import Path
from typing import Union, List

import torch
from torchvision.transforms import Compose, ToTensor, Normalize, Resize, InterpolationMode

from cn_clip.clip import _tokenizer
from cn_clip.clip.model import convert_weights, CLIP, restore_model

__all__ = ["load", "tokenize", "available_models", "image_transform", "load_from_name"]

_MODELSCOPE_ORG = "AI-ModelScope"
_HUGGINGFACE_ORG = "OFA-Sys"

_MODELS = {
    "ViT-B-16": ("chinese-clip-vit-base-patch16", "clip_cn_vit-b-16.pt"),
    "ViT-L-14": ("chinese-clip-vit-large-patch14", "clip_cn_vit-l-14.pt"),
    "ViT-L-14-336": ("chinese-clip-vit-large-patch14-336px", "clip_cn_vit-l-14-336.pt"),
    "ViT-H-14": ("chinese-clip-vit-huge-patch14", "clip_cn_vit-h-14.pt"),
    "RN50": ("chinese-clip-rn50", "clip_cn_rn50.pt"),
}

_MODEL_INFO = {
    "ViT-B-16": {
        "struct": "ViT-B-16@RoBERTa-wwm-ext-base-chinese",
        "input_resolution": 224
    },
    "ViT-L-14": {
        "struct": "ViT-L-14@RoBERTa-wwm-ext-base-chinese",
        "input_resolution": 224
    },
    "ViT-L-14-336": {
        "struct": "ViT-L-14-336@RoBERTa-wwm-ext-base-chinese",
        "input_resolution": 336
    },
    "ViT-H-14": {
        "struct": "ViT-H-14@RoBERTa-wwm-ext-large-chinese",
        "input_resolution": 224
    },
    "RN50": {
        "struct": "RN50@RBT3-chinese",
        "input_resolution": 224
    },
}


def _download(modelname: str, root: str, use_modelscope: bool = False):
    os.makedirs(root, exist_ok=True)

    # this a private function and the only caller has checked the model exsits
    reponame, filename = _MODELS[modelname]

    download_target = os.path.join(root, filename)

    if os.path.exists(download_target) and not os.path.isfile(download_target):
        raise RuntimeError(f"{download_target} exists and is not a regular file")

    if os.path.isfile(download_target):
        return download_target

    if use_modelscope:
        try:
            from modelscope.hub.file_download import model_file_download
        except ImportError as _:
            raise RuntimeError(
                "Ckpt download requires `modelscope`. "
                "Please install `modelscope` or download the ckpt manually and "
                "provide the local path so that we can continue."
            )

        local_path = model_file_download(
            model_id=_MODELSCOPE_ORG + "/" + reponame,
            file_path=filename,
            local_dir=root,
        )
    else:
        try:
            from huggingface_hub import hf_hub_download
        except ImportError as _:
            raise RuntimeError(
                "Ckpt download requires `huggingface_hub`. "
                "Please install `huggingface_hub` or download the ckpt manually and "
                "provide the local path so that we can continue."
            )

        local_path = hf_hub_download(
            _HUGGINGFACE_ORG + "/" + reponame, filename=filename, local_dir=root
        )

    return local_path


def _convert_image_to_rgb(image):
    return image.convert("RGB")


def available_models() -> List[str]:
    """Returns the names of available CLIP models"""
    return list(_MODELS.keys())


def load_from_name(name: str, device: Union[str, torch.device] = "cuda" if torch.cuda.is_available() else "cpu",
                   download_root: str = None, vision_model_name: str = None, text_model_name: str = None, input_resolution: int = None, use_modelscope: bool = False):
    if name in _MODELS:
        model_path = _download(name, download_root or os.path.expanduser("~/.cache/clip"), use_modelscope)
        model_name, model_input_resolution = _MODEL_INFO[name]['struct'], _MODEL_INFO[name]['input_resolution']
    elif os.path.isfile(name):
        assert vision_model_name and text_model_name and input_resolution, "Please specify specific 'vision_model_name', 'text_model_name', and 'input_resolution'"
        model_path = name
        model_name, model_input_resolution = f'{vision_model_name}@{text_model_name}', input_resolution
    else:
        raise RuntimeError(f"Model {name} not found; available models = {available_models()}")

    with open(model_path, 'rb') as opened_file:
        # loading saved checkpoint
        checkpoint = torch.load(opened_file, map_location="cpu")

    model = create_model(model_name, checkpoint)
    if str(device) == "cpu":
        model.float()
    else:
        model.to(device)
    return model, image_transform(model_input_resolution)


def load(model, device: Union[str, torch.device] = "cuda" if torch.cuda.is_available() else "cpu", clip_path=None,
         bert_path=None, use_flash_attention=False):
    """Load CLIP and BERT model weights
    """

    bert_state_dict = torch.load(bert_path, map_location="cpu") if bert_path else None
    clip_state_dict = torch.load(clip_path, map_location="cpu") if clip_path else None

    restore_model(model, clip_state_dict, bert_state_dict, use_flash_attention).to(device)

    if str(device) == "cpu":
        model.float()
    return model


def tokenize(texts: Union[str, List[str]], context_length: int = 52) -> torch.LongTensor:
    """
    Returns the tokenized representation of given input string(s)
    Parameters
    ----------
    texts : Union[str, List[str]]
        An input string or a list of input strings to tokenize
    context_length : int
        The context length to use; all baseline models use 52 as the context length
    Returns
    -------
    A two-dimensional tensor containing the resulting tokens, shape = [number of input strings, context_length]
    """
    if isinstance(texts, str):
        texts = [texts]

    all_tokens = []
    for text in texts:
        all_tokens.append([_tokenizer.vocab['[CLS]']] + _tokenizer.convert_tokens_to_ids(_tokenizer.tokenize(text))[
                                                        :context_length - 2] + [_tokenizer.vocab['[SEP]']])

    result = torch.zeros(len(all_tokens), context_length, dtype=torch.long)

    for i, tokens in enumerate(all_tokens):
        assert len(tokens) <= context_length
        result[i, :len(tokens)] = torch.tensor(tokens)

    return result


def _convert_to_rgb(image):
    return image.convert('RGB')


def image_transform(image_size=224):
    transform = Compose([
        Resize((image_size, image_size), interpolation=InterpolationMode.BICUBIC),
        _convert_to_rgb,
        ToTensor(),
        Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)),
    ])
    return transform


def create_model(model_name, checkpoint=None):
    vision_model, text_model = model_name.split('@')
    # Initialize the model.
    vision_model_config_file = Path(
        __file__).parent / f"model_configs/{vision_model.replace('/', '-')}.json"
    print('Loading vision model config from', vision_model_config_file)
    assert os.path.exists(vision_model_config_file)

    text_model_config_file = Path(
        __file__).parent / f"model_configs/{text_model.replace('/', '-')}.json"
    print('Loading text model config from', text_model_config_file)
    assert os.path.exists(text_model_config_file)

    with open(vision_model_config_file, 'r') as fv, open(text_model_config_file, 'r') as ft:
        model_info = json.load(fv)
        for k, v in json.load(ft).items():
            model_info[k] = v
    if isinstance(model_info['vision_layers'], str):
        model_info['vision_layers'] = eval(model_info['vision_layers'])
    print('Model info', model_info)
    model = CLIP(**model_info)
    convert_weights(model)
    if checkpoint:
        sd = checkpoint["state_dict"]
        if next(iter(sd.items()))[0].startswith('module'):
            sd = {k[len('module.'):]: v for k, v in sd.items() if "bert.pooler" not in k}
        model.load_state_dict(sd)
    return model



from collections import OrderedDict
from typing import Tuple, Union
from itertools import repeat
import collections.abc

import math
import logging
import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
from torch.utils.checkpoint import checkpoint

import importlib.util
if importlib.util.find_spec('flash_attn'):
    FlashMHA = importlib.import_module('flash_attn.flash_attention').FlashMHA

from cn_clip.clip import _tokenizer
from cn_clip.clip.configuration_bert import BertConfig
from cn_clip.clip.modeling_bert import BertModel


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1):
        super().__init__()

        # all conv layers have stride 1. an avgpool is performed after the second convolution when stride > 1
        self.conv1 = nn.Conv2d(inplanes, planes, 1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)

        self.conv2 = nn.Conv2d(planes, planes, 3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.avgpool = nn.AvgPool2d(stride) if stride > 1 else nn.Identity()

        self.conv3 = nn.Conv2d(planes, planes * self.expansion, 1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion)

        self.relu = nn.ReLU(inplace=True)
        self.downsample = None
        self.stride = stride

        if stride > 1 or inplanes != planes * Bottleneck.expansion:
            # downsampling layer is prepended with an avgpool, and the subsequent convolution has stride 1
            self.downsample = nn.Sequential(OrderedDict([
                ("-1", nn.AvgPool2d(stride)),
                ("0", nn.Conv2d(inplanes, planes * self.expansion, 1, stride=1, bias=False)),
                ("1", nn.BatchNorm2d(planes * self.expansion))
            ]))

    def forward(self, x: torch.Tensor):
        identity = x

        out = self.relu(self.bn1(self.conv1(x)))
        out = self.relu(self.bn2(self.conv2(out)))
        out = self.avgpool(out)
        out = self.bn3(self.conv3(out))

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)
        return out


class AttentionPool2d(nn.Module):
    def __init__(self, spacial_dim: int, embed_dim: int, num_heads: int, output_dim: int = None):
        super().__init__()
        self.positional_embedding = nn.Parameter(torch.randn(spacial_dim ** 2 + 1, embed_dim) / embed_dim ** 0.5)
        self.k_proj = nn.Linear(embed_dim, embed_dim)
        self.q_proj = nn.Linear(embed_dim, embed_dim)
        self.v_proj = nn.Linear(embed_dim, embed_dim)
        self.c_proj = nn.Linear(embed_dim, output_dim or embed_dim)
        self.num_heads = num_heads

    def forward(self, x):
        x = x.reshape(x.shape[0], x.shape[1], x.shape[2] * x.shape[3]).permute(2, 0, 1)  # NCHW -> (HW)NC
        x = torch.cat([x.mean(dim=0, keepdim=True), x], dim=0)  # (HW+1)NC
        x = x + self.positional_embedding[:, None, :].to(x.dtype)  # (HW+1)NC
        x, _ = F.multi_head_attention_forward(
            query=x, key=x, value=x,
            embed_dim_to_check=x.shape[-1],
            num_heads=self.num_heads,
            q_proj_weight=self.q_proj.weight,
            k_proj_weight=self.k_proj.weight,
            v_proj_weight=self.v_proj.weight,
            in_proj_weight=None,
            in_proj_bias=torch.cat([self.q_proj.bias, self.k_proj.bias, self.v_proj.bias]),
            bias_k=None,
            bias_v=None,
            add_zero_attn=False,
            dropout_p=0,
            out_proj_weight=self.c_proj.weight,
            out_proj_bias=self.c_proj.bias,
            use_separate_proj_weight=True,
            training=self.training,
            need_weights=False
        )

        return x[0]


class ModifiedResNet(nn.Module):
    """
    A ResNet class that is similar to torchvision's but contains the following changes:
    - There are now 3 "stem" convolutions as opposed to 1, with an average pool instead of a max pool.
    - Performs anti-aliasing strided convolutions, where an avgpool is prepended to convolutions with stride > 1
    - The final pooling layer is a QKV attention instead of an average pool
    """

    def __init__(self, layers, output_dim, heads, input_resolution=224, width=64):
        super().__init__()
        self.output_dim = output_dim
        self.input_resolution = input_resolution

        # the 3-layer stem
        self.conv1 = nn.Conv2d(3, width // 2, kernel_size=3, stride=2, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(width // 2)
        self.conv2 = nn.Conv2d(width // 2, width // 2, kernel_size=3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(width // 2)
        self.conv3 = nn.Conv2d(width // 2, width, kernel_size=3, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(width)
        self.avgpool = nn.AvgPool2d(2)
        self.relu = nn.ReLU(inplace=True)

        # residual layers
        self._inplanes = width  # this is a *mutable* variable used during construction
        self.layer1 = self._make_layer(width, layers[0])
        self.layer2 = self._make_layer(width * 2, layers[1], stride=2)
        self.layer3 = self._make_layer(width * 4, layers[2], stride=2)
        self.layer4 = self._make_layer(width * 8, layers[3], stride=2)

        embed_dim = width * 32  # the ResNet feature dimension
        self.attnpool = AttentionPool2d(input_resolution // 32, embed_dim, heads, output_dim)

    def _make_layer(self, planes, blocks, stride=1):
        layers = [Bottleneck(self._inplanes, planes, stride)]

        self._inplanes = planes * Bottleneck.expansion
        for _ in range(1, blocks):
            layers.append(Bottleneck(self._inplanes, planes))

        return nn.Sequential(*layers)

    @torch.jit.ignore
    def set_grad_checkpointing(self, enable=True):
        # FIXME support for non-transformer
        pass

    def forward(self, x):
        def stem(x):
            for conv, bn in [(self.conv1, self.bn1), (self.conv2, self.bn2), (self.conv3, self.bn3)]:
                x = self.relu(bn(conv(x)))
            x = self.avgpool(x)
            return x

        x = x.type(self.conv1.weight.dtype)
        x = stem(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.attnpool(x)

        return x


class LayerNorm(nn.LayerNorm):
    """Subclass torch's LayerNorm to handle fp16."""

    def forward(self, x: torch.Tensor):
        orig_type = x.dtype
        ret = super().forward(x.type(torch.float32))
        return ret.type(orig_type)


class QuickGELU(nn.Module):
    def forward(self, x: torch.Tensor):
        return x * torch.sigmoid(1.702 * x)


class ResidualAttentionBlock(nn.Module):
    def __init__(self, d_model: int, n_head: int, attn_mask: torch.Tensor = None, use_flash_attention: bool = False):
        super().__init__()

        self.attn = nn.MultiheadAttention(d_model, n_head) if not use_flash_attention else FlashMHA(d_model, n_head)
        self.ln_1 = LayerNorm(d_model)
        self.mlp = nn.Sequential(OrderedDict([
            ("c_fc", nn.Linear(d_model, d_model * 4)),
            ("gelu", QuickGELU()),
            ("c_proj", nn.Linear(d_model * 4, d_model))
        ]))
        self.ln_2 = LayerNorm(d_model)
        self.attn_mask = attn_mask
        self.use_flash_attention = use_flash_attention

    def attention(self, x: torch.Tensor):
        self.attn_mask = self.attn_mask.to(dtype=x.dtype, device=x.device) if self.attn_mask is not None else None
        if self.use_flash_attention:
            # Batch first is needed for FlashAttention. See https://github.com/HazyResearch/flash-attention/issues/84 for more information.
            return self.attn(x.transpose(1, 0))[0].transpose(1, 0)
        else:
            return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x: torch.Tensor):
        x = x + self.attention(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class Transformer(nn.Module):
    def __init__(self, width: int, layers: int, heads: int, attn_mask: torch.Tensor = None, use_flash_attention: bool = False):
        super().__init__()
        self.width = width
        self.layers = layers
        self.grad_checkpointing = False
        self.resblocks = nn.Sequential(*[ResidualAttentionBlock(width, heads, attn_mask, use_flash_attention) for _ in range(layers)])

    def forward(self, x: torch.Tensor):
        if self.grad_checkpointing and not torch.jit.is_scripting():
            for r in self.resblocks:
                x = checkpoint(r, x)
            return x        
        return self.resblocks(x)


class VisualTransformer(nn.Module):
    def __init__(self, input_resolution: int, patch_size: int, width: int, layers: int, heads: int, output_dim: int, use_flash_attention: bool = False):
        super().__init__()
        self.input_resolution = input_resolution
        self.grid_size = (self.input_resolution // patch_size, self.input_resolution // patch_size)
        self.output_dim = output_dim
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=width, kernel_size=patch_size, stride=patch_size, bias=False)

        scale = width ** -0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(width))
        self.positional_embedding = nn.Parameter(scale * torch.randn((input_resolution // patch_size) ** 2 + 1, width))
        self.ln_pre = LayerNorm(width)

        self.transformer = Transformer(width, layers, heads, use_flash_attention=use_flash_attention)

        self.ln_post = LayerNorm(width)
        self.proj = nn.Parameter(scale * torch.randn(width, output_dim))

    @torch.jit.ignore
    def set_grad_checkpointing(self, enable=True):
        self.transformer.grad_checkpointing = enable

    def random_masking(self, x, mask_ratio):
        N, L, D = x.shape  # batch, length, dim
        len_keep = int((L - 1) * (1 - mask_ratio))

        noise = torch.rand(N, L - 1, device=x.device)
        ids_shuffle = torch.argsort(noise, dim=1) + torch.ones(N, L - 1, device=x.device,
                                                               dtype=int)
        ids_keep = ids_shuffle[:, :len_keep]

        x_masked = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, D))

        x0 = x[:, 0, :]
        x0 = x0.reshape(N, 1, D)
        x_masked_add = torch.cat([x0, x_masked], axis=1)
        return x_masked_add

    def forward(self, x: torch.Tensor, mask_ratio: float = 0.0):
        x = self.conv1(x)  # shape = [*, width, grid, grid]
        x = x.reshape(x.shape[0], x.shape[1], -1)  # shape = [*, width, grid ** 2]
        x = x.permute(0, 2, 1)  # shape = [*, grid ** 2, width]
        x = torch.cat([self.class_embedding.to(x.dtype) + torch.zeros(x.shape[0], 1, x.shape[-1], dtype=x.dtype, device=x.device), x], dim=1)  # shape = [*, grid ** 2 + 1, width]
        x = x + self.positional_embedding.to(x.dtype)
        if mask_ratio != 0:
            x = self.random_masking(x, mask_ratio)
        x = self.ln_pre(x)

        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD

        x = self.ln_post(x[:, 0, :])

        if self.proj is not None:
            x = x @ self.proj

        return x


class CLIP(nn.Module):
    def __init__(self,
                 embed_dim: int,
                 # vision
                 image_resolution: int,
                 vision_layers: Union[Tuple[int, int, int, int], int],
                 vision_width: int,
                 vision_patch_size: int,
                 # text
                 vocab_size: int,
                 text_attention_probs_dropout_prob: float, 
                 text_hidden_act: str, 
                 text_hidden_dropout_prob: float, 
                 text_hidden_size: int,
                 text_initializer_range: float, 
                 text_intermediate_size: int, 
                 text_max_position_embeddings: int, 
                 text_num_attention_heads: int, 
                 text_num_hidden_layers: int, 
                 text_type_vocab_size: int,
                 tokenizer = _tokenizer,
                 # vision head width, added this param for ViT-H
                 vision_head_width: int = 64,
                 use_flash_attention: bool = False,
                 ):
        super().__init__()

        if isinstance(vision_layers, (tuple, list)):
            vision_heads = vision_width * 32 // vision_head_width
            self.visual = ModifiedResNet(
                layers=vision_layers,
                output_dim=embed_dim,
                heads=vision_heads,
                input_resolution=image_resolution,
                width=vision_width
            )
        else:
            vision_heads = vision_width // vision_head_width
            self.visual = VisualTransformer(
                input_resolution=image_resolution,
                patch_size=vision_patch_size,
                width=vision_width,
                layers=vision_layers,
                heads=vision_heads,
                output_dim=embed_dim,
                use_flash_attention=use_flash_attention
            )

        self.bert_config = BertConfig(
            vocab_size_or_config_json_file=vocab_size,
            hidden_size=text_hidden_size,
            num_hidden_layers=text_num_hidden_layers,
            num_attention_heads=text_num_attention_heads,
            intermediate_size=text_intermediate_size,
            hidden_act=text_hidden_act,
            hidden_dropout_prob=text_hidden_dropout_prob,
            attention_probs_dropout_prob=text_attention_probs_dropout_prob,
            max_position_embeddings=text_max_position_embeddings,
            type_vocab_size=text_type_vocab_size,
            initializer_range=text_initializer_range,
            layer_norm_eps=1e-12,
            use_flash_attention=use_flash_attention
        )
        self.bert = BertModel(self.bert_config)

        self.text_projection = nn.Parameter(torch.empty(text_hidden_size, embed_dim))
        self.logit_scale = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

        self.tokenizer = tokenizer

        self.initialize_parameters()

    def initialize_parameters(self):
        self.logit_scale = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

        if isinstance(self.visual, ModifiedResNet):
            if self.visual.attnpool is not None:
                std = self.visual.attnpool.c_proj.in_features ** -0.5
                nn.init.normal_(self.visual.attnpool.q_proj.weight, std=std)
                nn.init.normal_(self.visual.attnpool.k_proj.weight, std=std)
                nn.init.normal_(self.visual.attnpool.v_proj.weight, std=std)
                nn.init.normal_(self.visual.attnpool.c_proj.weight, std=std)

            for resnet_block in [self.visual.layer1, self.visual.layer2, self.visual.layer3, self.visual.layer4]:
                for name, param in resnet_block.named_parameters():
                    if name.endswith("bn3.weight"):
                        nn.init.zeros_(param)

        if self.text_projection is not None:
            nn.init.normal_(self.text_projection, std=self.bert_config.hidden_size ** -0.5)

    @torch.jit.ignore
    def set_grad_checkpointing(self, enable=True):
        self.visual.set_grad_checkpointing(enable)
        self.bert.set_grad_checkpointing(enable)

    @property
    def dtype(self):
        return self.visual.conv1.weight.dtype

    def encode_image(self, image, mask_ratio=0):
        if isinstance(self.visual, ModifiedResNet):
            # mask_ratio > 0 (FLIP strategy) is currently only implemented for VisualTransformer.
            return self.visual(image.type(self.dtype))
        return self.visual(image.type(self.dtype), mask_ratio)

    def encode_text(self, text):
        pad_index = self.tokenizer.vocab['[PAD]']
        attn_mask = text.ne(pad_index).type(self.dtype)
        x = self.bert(text, attention_mask=attn_mask)[0].type(self.dtype) # [batch_size, seq_length, hidden_size]
        return x[:, 0, :] @ self.text_projection

    def forward(self, image, text, mask_ratio=0):
        assert image is not None or text is not None, "text and image cannot both be None!"

        if image is None:
            return self.encode_text(text)
        elif text is None:
            return self.encode_image(image)
        image_features = self.encode_image(image, mask_ratio)
        text_features = self.encode_text(text)

        image_features = image_features / image_features.norm(dim=-1, keepdim=True)
        text_features = text_features / text_features.norm(dim=-1, keepdim=True)

        return image_features, text_features, self.logit_scale.exp()

    def get_similarity(self, image, text):
        image_features = self.encode_image(image)
        text_features = self.encode_text(text)

        # normalized features
        image_features = image_features / image_features.norm(dim=1, keepdim=True)
        text_features = text_features / text_features.norm(dim=1, keepdim=True)

        # cosine similarity as logits
        logit_scale = self.logit_scale.exp()
        logits_per_image = logit_scale * image_features @ text_features.t()
        logits_per_text = logits_per_image.t()

        # shape = [global_batch_size, global_batch_size]
        return logits_per_image, logits_per_text


def convert_models_to_fp32(model):
    for p in model.parameters():
        p.data = p.data.float()
        if p.grad:
            p.grad.data = p.grad.data.float()


def convert_weights(model: nn.Module):
    """Convert applicable model parameters to fp16"""

    def _convert_weights_to_fp16(l):
        if isinstance(l, (nn.Conv1d, nn.Conv2d, nn.Linear)):
            l.weight.data = l.weight.data.half()
            if l.bias is not None:
                l.bias.data = l.bias.data.half()

        if isinstance(l, nn.MultiheadAttention):
            for attr in [*[f"{s}_proj_weight" for s in ["in", "q", "k", "v"]], "in_proj_bias", "bias_k", "bias_v"]:
                tensor = getattr(l, attr)
                if tensor is not None:
                    tensor.data = tensor.data.half()

        if isinstance(l, BertModel):
            l.to(torch.half)

        for name in ["text_projection", "proj"]:
            if hasattr(l, name):
                attr = getattr(l, name)
                if attr is not None:
                    attr.data = attr.data.half()

    model.apply(_convert_weights_to_fp16)


def restore_model(model, clip_state_dict: dict, bert_state_dict: dict, use_flash_attention: bool):
    merged_state_dict = {}

    # use clip_state_dict to initialize the image encoder & logit scale
    if clip_state_dict is not None:
        for k, v in clip_state_dict.items():
            if k.startswith("visual") or k == "logit_scale":
                merged_state_dict[k] = v

    # use bert_state_dict to initialize the text encoder
    if bert_state_dict is not None:
        for k, v in bert_state_dict.items():
            if k.startswith("bert") and "bert.pooler" not in k:
                merged_state_dict[k] = v

    # adapt flash attention
    if use_flash_attention:
        merged_state_dict = convert_state_dict(merged_state_dict)

    convert_weights(model)
    resize_pos_embed(merged_state_dict, model)
    model.load_state_dict(merged_state_dict, strict=False)
    return model.eval()


def convert_state_dict(state_dict):
    """Adapt to Flash Attention"""
    if not state_dict:
        return state_dict

    prefix = 'module.' if list(state_dict.keys())[0].startswith('module') else ''

    if f'{prefix}visual.transformer.resblocks.0.attn.in_proj_weight' in state_dict:
        for k in list(state_dict.keys()):
            if 'attn.in_proj_weight' in k:
                state_dict[k.replace('attn.in_proj_weight', 'attn.Wqkv.weight')] = state_dict.pop(k)
            elif 'attn.in_proj_bias' in k:
                state_dict[k.replace('attn.in_proj_bias', 'attn.Wqkv.bias')] = state_dict.pop(k)
    elif f'{prefix}visual.transformer.resblocks.0.attn.Wqkv.weight' in state_dict:
        for k in list(state_dict.keys()):
            if 'attn.Wqkv.weight' in k:
                state_dict[k.replace('attn.Wqkv.weight', 'attn.in_proj_weight')] = state_dict.pop(k)
            elif 'attn.Wqkv.bias' in k:
                state_dict[k.replace('attn.Wqkv.bias', 'attn.in_proj_bias')] = state_dict.pop(k)

    if f'{prefix}bert.encoder.layer.0.attention.self.query.weight' in state_dict:
        i = 0
        while f'{prefix}bert.encoder.layer.{i}.attention.self.query.weight' in state_dict:
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.Wqkv.weight'] = torch.cat(
                (state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.query.weight'),
                 state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.key.weight'),
                 state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.value.weight'))
            )
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.Wqkv.bias'] = torch.cat(
                (state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.query.bias'),
                 state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.key.bias'),
                 state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.value.bias'))
            )
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.out_proj.weight'] = \
                state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.output.dense.weight')
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.out_proj.bias'] = \
                state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.output.dense.bias')
            i += 1
    elif f'{prefix}bert.encoder.layer.0.attention.self.Wqkv.weight' in state_dict:
        i = 0
        while f'{prefix}bert.encoder.layer.{i}.attention.self.Wqkv.weight' in state_dict:
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.query.weight'], \
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.key.weight'], \
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.value.weight'] = \
                torch.chunk(state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.Wqkv.weight'), chunks=3)
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.query.bias'], \
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.key.bias'], \
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.self.value.bias'] = \
                torch.chunk(state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.Wqkv.bias'), chunks=3)
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.output.dense.weight'] = \
                state_dict.pop(f'{prefix}bert.encoder.layer.{i}.attention.self.out_proj.weight')
            state_dict[f'{prefix}bert.encoder.layer.{i}.attention.output.dense.bias'] = \
                state_dict.pop(f'module.bert.encoder.layer.{i}.attention.self.out_proj.bias')
            i += 1

    return state_dict


def resize_pos_embed(state_dict, model, interpolation: str = 'bicubic', seq_dim=1, prefix=""):
    # Rescale the grid of position embeddings when loading from state_dict
    old_pos_embed = state_dict.get(prefix + 'visual.positional_embedding', None)
    model = model.module if hasattr(model, 'module') else model
    if old_pos_embed is None or not hasattr(model.visual, 'grid_size'):
        return
    grid_size = to_2tuple(model.visual.grid_size)
    extra_tokens = 1  # FIXME detect different token configs (ie no class token, or more)
    new_seq_len = grid_size[0] * grid_size[1] + extra_tokens
    if new_seq_len == old_pos_embed.shape[0]:
        return

    if extra_tokens:
        pos_emb_tok, pos_emb_img = old_pos_embed[:extra_tokens], old_pos_embed[extra_tokens:]
    else:
        pos_emb_tok, pos_emb_img = None, old_pos_embed
    old_grid_size = to_2tuple(int(math.sqrt(len(pos_emb_img))))

    logging.info('Resizing position embedding grid-size from %s to %s', old_grid_size, grid_size)
    pos_emb_img = pos_emb_img.reshape(1, old_grid_size[0], old_grid_size[1], -1).permute(0, 3, 1, 2)
    pos_emb_img = F.interpolate(
        pos_emb_img,
        size=grid_size,
        mode=interpolation,
        align_corners=True,
    )
    pos_emb_img = pos_emb_img.permute(0, 2, 3, 1).reshape(1, grid_size[0] * grid_size[1], -1)[0]
    if pos_emb_tok is not None:
        new_pos_embed = torch.cat([pos_emb_tok, pos_emb_img], dim=0)
    else:
        new_pos_embed = pos_emb_img
    state_dict[prefix + 'visual.positional_embedding'] = new_pos_embed


# From PyTorch internals
def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse


to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)
to_3tuple = _ntuple(3)
to_4tuple = _ntuple(4)
to_ntuple = lambda n, x: _ntuple(n)(x)



# -*- coding: utf-8 -*-
"""
This script converts PyTorch implemented Chinese-CLIP (text or vision) model to CoreML format for deployment in Apple's ecosystem.
"""

import os
import argparse
from PIL import Image
import torch
from torch import nn
import coremltools as ct
import cn_clip.clip as clip
from cn_clip.clip.utils import _MODELS, _MODEL_INFO, _download, available_models, create_model, image_transform


class ImageEncoder(nn.Module):
    def __init__(self, clip_model):
        super().__init__()
        self.clip_model = clip_model

    def forward(self, image):
        return self.clip_model.encode_image(image)


class TextEncoder(nn.Module):
    def __init__(self, clip_model):
        super().__init__()
        self.clip_model = clip_model

    def forward(self, text):
        return self.clip_model.encode_text(text)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-arch",
        required=True,
        choices=["ViT-B-16", "ViT-L-14", "ViT-L-14-336", "ViT-H-14", "RN50"],
        help="Specify the architecture (model scale) of Chinese-CLIP model to be converted."
    )
    parser.add_argument(
        "--pytorch-ckpt-path",
        default=None,
        type=str,
        help="Path of the input PyTorch Chinese-CLIP checkpoint."
    )
    parser.add_argument(
        "--download-root",
        default=None,
        type=str,
        help="If --pytorch-ckpt-path is None, official pretrained ckpt will be downloaded under --download-root directory and converted."
    )
    parser.add_argument(
        "--save-coreml-path",
        required=True,
        type=str,
        help="Path (prefix) of the output converted CoreML Chinese-CLIP text or vision model."
    )
    parser.add_argument(
        "--convert-text",
        action="store_true",
        help="Whether to convert the text encoder (text feature extractor) into CoreML."
    )
    parser.add_argument(
        "--convert-vision",
        action="store_true",
        help="Whether to convert the vision encoder (vision feature extractor) into CoreML."
    )
    parser.add_argument(
        "--precision",
        default="fp16",
        choices=["fp16", "fp32"],
        help="Specify the architecture (model scale) of Chinese-CLIP model to be converted."
    )
    parser.add_argument(
        "--context-length", type=int, default=52, help="The padded length of input text (include [CLS] & [SEP] tokens)."
    )
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()

    # Log params.
    print("Params:")
    for name in sorted(vars(args)):
        val = getattr(args, name)
        print(f"  {name}: {val}")

    # prepare the PyTorch model weights
    if os.path.isfile(args.pytorch_ckpt_path):
        input_ckpt_path = args.pytorch_ckpt_path
    elif args.model_arch in _MODELS:
        input_ckpt_path = _download(
            _MODELS[args.model_arch], args.download_root or os.path.expanduser("./cache/clip"))
    else:
        raise RuntimeError(
            f"Model {args.model_arch} not found; available models = {available_models()}")

    with open(input_ckpt_path, 'rb') as opened_file:
        checkpoint = torch.load(opened_file, map_location="cpu")

    # prepare the PyTorch implemented model and restore weights
    model = create_model(
        _MODEL_INFO[args.model_arch]['struct'], checkpoint).float().eval()

    # prepare empty image and text as input placeholders for CoreML
    resolution = _MODEL_INFO[args.model_arch]['input_resolution']
    preprocess = image_transform(resolution)
    if args.precision == "fp16":
        precision = ct.precision.FLOAT16
    elif args.precision == "fp32":
        precision = ct.precision.FLOAT32
    image = preprocess(Image.new('RGB', (resolution, resolution))).unsqueeze(0)
    text = clip.tokenize([""], context_length=args.context_length)

    # perform conversions, CoreML text and vision encoders will be saved into separated files
    if args.convert_text:
        # Prepare the model for conversion
        text_model = TextEncoder(model)
        text_model.eval()

        # Prepare text input
        text = clip.tokenize([""], context_length=args.context_length).int()

        # Trace the model for text input
        traced_text_model = torch.jit.trace(text_model, text)

        # Convert traced model to CoreML
        text_outputs = [ct.TensorType(
            name="text_features")]
        text_coreml_model = ct.convert(
            traced_text_model,
            inputs=[ct.TensorType(name="text", shape=text.shape)],
            outputs=text_outputs,
            convert_to="mlprogram",
            compute_precision=precision,
            minimum_deployment_target=ct.target.iOS15
        )

        # Save the CoreML model
        text_coreml_model_path = f"{args.save_coreml_path}.text.mlpackage"
        text_coreml_model.save(text_coreml_model_path)
        print(
            f"Text model converted to CoreML and saved at: {text_coreml_model_path}")

    if args.convert_vision:
        # Prepare the model for conversion
        image_model = ImageEncoder(model)
        image_model.eval()

        # Prepare a dummy image input
        image_width = 336 if args.model_arch == "ViT-L-14-336" else 224
        dummy_image_input = torch.rand(1, 3, image_width, image_width)

        # Trace the model for image input
        traced_image_model = torch.jit.trace(image_model, dummy_image_input)

        # Convert traced model to CoreML
        image_outputs = [ct.TensorType(name="image_features")]
        image_coreml_model = ct.convert(
            traced_image_model,
            inputs=[ct.TensorType(
                name="image", shape=dummy_image_input.shape)],
            outputs=image_outputs,
            convert_to="mlprogram",
            compute_precision=precision,
            minimum_deployment_target=ct.target.iOS15
        )

        # Save the CoreML model
        image_coreml_model_path = f"{args.save_coreml_path}.image.mlpackage"
        image_coreml_model.save(image_coreml_model_path)
        print(
            f"Image model converted to CoreML and saved at: {image_coreml_model_path}")



# -*- coding: utf-8 -*-
"""
The util functions used in inference speed comparison.
"""

from typing import  List
import numpy as np
from contextlib import contextmanager
import time

def print_timings(name: str, timings: List[float]) -> None:
    """
    Format and print inference latencies.
    :param name: inference engine name
    :param timings: latencies measured during the inference
    """
    mean_time = 1e3 * np.mean(timings)
    std_time = 1e3 * np.std(timings)
    min_time = 1e3 * np.min(timings)
    max_time = 1e3 * np.max(timings)
    median, percent_95_time, percent_99_time = 1e3 * np.percentile(timings, [50, 95, 99])
    print(
        f"[{name}] "
        f"mean={mean_time:.2f}ms, "
        f"sd={std_time:.2f}ms, "
        f"min={min_time:.2f}ms, "
        f"max={max_time:.2f}ms, "
        f"median={median:.2f}ms, "
        f"95p={percent_95_time:.2f}ms, "
        f"99p={percent_99_time:.2f}ms"
    )


@contextmanager
def track_infer_time(buffer: List[int]) -> None:
    """
    A context manager to perform latency measures
    :param buffer: a List where to save latencies for each input
    """
    start = time.perf_counter()
    yield
    end = time.perf_counter()
    buffer.append(end - start)


# -*- coding: utf-8 -*-
"""
This script converts PyTorch implemented Chinese-CLIP (text or vision) model to ONNX format for CPU/GPU deployment.
"""

import os
import argparse
from PIL import Image
import torch
import torch.onnx
from onnx import load_model, save_model
from onnxmltools.utils import convert_float_to_float16
import cn_clip.clip as clip
from clip.utils import _MODELS, _MODEL_INFO, _download, available_models, create_model, image_transform

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-arch", 
        required=True, 
        choices=["ViT-B-16", "ViT-L-14", "ViT-L-14-336", "ViT-H-14", "RN50"],
        help="Specify the architecture (model scale) of Chinese-CLIP model to be converted."
    )
    parser.add_argument(
        "--pytorch-ckpt-path", 
        default=None, 
        type=str, 
        help="Path of the input PyTorch Chinese-CLIP checkpoint. Default to None which will automatically download the pretrained checkpoint."
    )
    parser.add_argument(
        "--download-root", 
        default=None, 
        type=str, 
        help="If --pytorch-ckpt-path is None, official pretrained ckpt will be downloaded under --download-root directory and converted. Default to ~/cache/clip/ ."
    )
    parser.add_argument(
        "--save-onnx-path", 
        required=True,
        type=str, 
        help="Path (prefix) of the output converted ONNX Chinese-CLIP text or vision model."
    )
    parser.add_argument(
        "--convert-text",
        action="store_true",
        help="Whether to convert the text encoder (text feature extractor) into ONNX."
    )
    parser.add_argument(
        "--convert-vision",
        action="store_true",
        help="Whether to convert the vision encoder (vision feature extractor) into ONNX."
    )
    parser.add_argument(
        "--context-length", type=int, default=52, help="The padded length of input text (include [CLS] & [SEP] tokens). Default to 52."
    )
    args = parser.parse_args()
    return args


def packing_small_onnx_files(onnx_path):
    # packing small files into an extra file
    save_model(load_model(onnx_path), 
            onnx_path, 
            location="{}.extra_file".format(os.path.split(onnx_path)[1]),
            save_as_external_data=True,
            all_tensors_to_one_file=True,
            size_threshold=1024,
            convert_attribute=True)
    # remove small files
    onnx_dir = os.path.split(onnx_path)[0]
    for key in checkpoint['state_dict']:
        if key.startswith('module.visual'):
            small_file_path = os.path.join(onnx_dir, key[7:])
            if os.path.exists(small_file_path):
                os.remove(small_file_path)
        if key.startswith('visual'):
            small_file_path = os.path.join(onnx_dir, key)
            if os.path.exists(small_file_path):
                os.remove(small_file_path)                    
    os.system("rm -f {}".format(os.path.join(onnx_dir, "Constant_*_attr__value")))


if __name__ == '__main__':
    args = parse_args()

    # Log params.
    print("Params:")
    for name in sorted(vars(args)):
        val = getattr(args, name)
        print(f"  {name}: {val}")

    # prepare the PyTorch model weights
    if os.path.isfile(args.pytorch_ckpt_path):
        input_ckpt_path = args.pytorch_ckpt_path
    elif args.model_arch in _MODELS:
        input_ckpt_path = _download(_MODELS[args.model_arch], args.download_root or os.path.expanduser("./cache/clip"))
    else:
        raise RuntimeError(f"Model {args.model_arch} not found; available models = {available_models()}")

    with open(input_ckpt_path, 'rb') as opened_file:
        checkpoint = torch.load(opened_file, map_location="cpu")

    # prepare the PyTorch implemented model and restore weights
    model = create_model(_MODEL_INFO[args.model_arch]['struct'], checkpoint).float().eval()

    # prepare empty image and text as input placeholders for ONNX
    resolution = _MODEL_INFO[args.model_arch]['input_resolution']
    preprocess = image_transform(resolution)
    image = preprocess(Image.new('RGB', (resolution, resolution))).unsqueeze(0)
    text = clip.tokenize([""], context_length=args.context_length)

    # perform conversions, ONNX text and vision encoders will be saved into separated files
    if args.convert_text:
        # convert text FP32 ONNX model
        text_fp32_onnx_path = f"{args.save_onnx_path}.txt.fp32.onnx"
        torch.onnx.export(model,
                    (None, text),
                    text_fp32_onnx_path,
                    input_names=['text'],
                    output_names=['unnorm_text_features'],
                    export_params=True,
                    opset_version=13,
                    verbose=True)
        # convert text FP16 ONNX model based on the FP32 model
        text_fp16_onnx_path = f"{args.save_onnx_path}.txt.fp16.onnx"
        text_fp32_onnx_model = load_model(text_fp32_onnx_path)
        text_fp16_onnx_model = convert_float_to_float16(text_fp32_onnx_model, keep_io_types=True, disable_shape_infer=True)
        save_model(text_fp16_onnx_model,
                    text_fp16_onnx_path,
                    location="{}.extra_file".format(os.path.split(text_fp16_onnx_path)[1]),
                    save_as_external_data=True,
                    all_tensors_to_one_file=True,
                    size_threshold=1024,
                    convert_attribute=True)

    if args.convert_vision:
        # convert vision FP32 ONNX model
        vision_fp32_onnx_path = f"{args.save_onnx_path}.img.fp32.onnx"
        vision_fp32_onnx_hasextra = False
        torch.onnx.export(model,
                    (image, None),
                    vision_fp32_onnx_path,
                    input_names=['image'],
                    output_names=['unnorm_image_features'],
                    export_params=True,
                    do_constant_folding=False,
                    opset_version=13,
                    verbose=True)
        # for ViT-H-14 FP32 model, make another conversion to deal with the generated small files
        if args.model_arch == "ViT-H-14":
            packing_small_onnx_files(vision_fp32_onnx_path)
            vision_fp32_onnx_hasextra = True
        # convert vision FP16 ONNX model based on the FP32 model
        vision_fp16_onnx_path = f"{args.save_onnx_path}.img.fp16.onnx"
        vision_fp32_onnx_model = load_model(vision_fp32_onnx_path)
        vision_fp16_onnx_model = convert_float_to_float16(vision_fp32_onnx_model, keep_io_types=True, disable_shape_infer=True)
        save_model(vision_fp16_onnx_model,
                    vision_fp16_onnx_path,
                    location="{}.extra_file".format(os.path.split(vision_fp16_onnx_path)[1]),
                    save_as_external_data=True,
                    all_tensors_to_one_file=True,
                    size_threshold=1024,
                    convert_attribute=True)

    print("Finished PyTorch to ONNX conversion...")
    if args.convert_text:
        print(f">>> The text FP32 ONNX model is saved at {text_fp32_onnx_path}")
        print(f">>> The text FP16 ONNX model is saved at {text_fp16_onnx_path} with extra file {text_fp16_onnx_path}.extra_file")
    if args.convert_vision:
        print(f">>> The vision FP32 ONNX model is saved at {vision_fp32_onnx_path}" + \
            (f" with extra file {vision_fp32_onnx_path}.extra_file" if vision_fp32_onnx_hasextra else ""))
        print(f">>> The vision FP16 ONNX model is saved at {vision_fp16_onnx_path} with extra file {vision_fp16_onnx_path}.extra_file")


import os
import time
import json
import logging
import numpy as np
from tqdm import tqdm

import torch
import torch.nn as nn
from torch.cuda.amp import autocast
import torch.distributed.nn
import torch.distributed as dist
import torch.nn.functional as F

from cn_clip.clip.model import convert_state_dict


def is_master(args):
    return args.rank == 0

def get_loss(model, images, texts, loss_img, loss_txt, args, accum_image_features=None, accum_text_features=None, accum_idx=-1, teacher_model=None, teacher_accum_image_features=None):
    if args.accum_freq == 1:
        image_features, text_features, logit_scale = model(images, texts, args.mask_ratio)

        if args.distillation:
            with torch.no_grad():
                # different teacher model has different output
                output = teacher_model.module.get_feature(images)
                if(isinstance(output, tuple)):
                    teacher_image_features = output[0]
                else:
                    teacher_image_features = output
    else:
        assert accum_image_features and accum_text_features and accum_idx != -1
        chunk_image_features, chunk_text_features, logit_scale = model(images, texts, args.mask_ratio)

        if args.distillation:
            with torch.no_grad():
                # different teacher model has different output
                output = teacher_model.module.get_feature(images)
                if(isinstance(output, tuple)):
                    teacher_chunk_image_features = output[0]
                else:
                    teacher_chunk_image_features = output
            teacher_image_features = torch.cat(
            teacher_accum_image_features[:accum_idx] + [teacher_chunk_image_features] + teacher_accum_image_features[accum_idx + 1:])
        
        image_features = torch.cat(
            accum_image_features[:accum_idx] + [chunk_image_features] + accum_image_features[accum_idx + 1:])
        text_features = torch.cat(
            accum_text_features[:accum_idx] + [chunk_text_features] + accum_text_features[accum_idx + 1:])
    logit_scale = logit_scale.mean()
    if args.aggregate:
        world_size = dist.get_world_size()
        rank = dist.get_rank()

        # We gather tensors from all gpus to get more negatives to contrast with.
        if args.gather_with_grad:
            all_image_features = torch.cat(torch.distributed.nn.all_gather(image_features), dim=0)
            all_text_features = torch.cat(torch.distributed.nn.all_gather(text_features), dim=0)

            if args.distillation:
                all_teacher_image_features = torch.cat(torch.distributed.nn.all_gather(teacher_image_features), dim=0)
        else:
            gathered_image_features = [
                torch.zeros_like(image_features) for _ in range(world_size)
            ]
            gathered_text_features = [
                torch.zeros_like(text_features) for _ in range(world_size)
            ]
            
            dist.all_gather(gathered_image_features, image_features)
            dist.all_gather(gathered_text_features, text_features)

            all_image_features = torch.cat(
                [image_features]
                + gathered_image_features[:rank]
                + gathered_image_features[rank + 1 :]
            )
            all_text_features = torch.cat(
                [text_features]
                + gathered_text_features[:rank]
                + gathered_text_features[rank + 1 :]
            )

        # this is needed to send gradients back everywhere.
        logits_per_image = logit_scale * all_image_features @ all_text_features.t()
        logits_per_text = logits_per_image.t()

        if args.distillation:
            gathered_teacher_image_features = [
                torch.zeros_like(teacher_image_features) for _ in range(world_size)
            ]
            dist.all_gather(gathered_teacher_image_features, teacher_image_features)
            all_teacher_image_features = torch.cat(
                [teacher_image_features]
                + gathered_teacher_image_features[:rank]
                + gathered_teacher_image_features[rank + 1 :]
            )
            kd_loss = cosineSimilarityLoss(all_teacher_image_features, all_image_features)

    else:
        logits_per_image = logit_scale * image_features @ text_features.t()
        logits_per_text = logit_scale * text_features @ image_features.t()

        if args.distillation:
            kd_loss = cosineSimilarityLoss(teacher_image_features, image_features)

    ground_truth = torch.arange(len(logits_per_image)).long()
    ground_truth = ground_truth.cuda(args.local_device_rank, non_blocking=True)

    total_loss = (
        loss_img(logits_per_image, ground_truth)
        + loss_txt(logits_per_text, ground_truth)
    ) / 2

    acc = None
    if args.report_training_batch_acc:
        i2t_acc = (logits_per_image.argmax(-1) == ground_truth).sum() / len(logits_per_image)
        t2i_acc = (logits_per_text.argmax(-1) == ground_truth).sum() / len(logits_per_text)
        acc = {"i2t": i2t_acc, "t2i": t2i_acc}

    if args.distillation:
        total_loss += kd_loss * args.kd_loss_weight

    return total_loss, acc

def freeze_vision_bn(args, model):
    # freeze bn running mean and variance
    if 'RN' in args.vision_model:
        RN_visual_modules = model.module.visual.modules() if isinstance(model, nn.parallel.DistributedDataParallel) else model.visual.modules()
        for m in RN_visual_modules:
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

def train(model, data, epoch, optimizer, scaler, scheduler, args, global_trained_steps, teacher_model=None):
    # os.environ["WDS_EPOCH"] = str(epoch)

    model.train()
    if args.freeze_vision:
        freeze_vision_bn(args, model)

    dataloader, sampler = data['train'].dataloader,  data['train'].sampler

    loss_img = nn.CrossEntropyLoss()
    loss_txt = nn.CrossEntropyLoss()

    loss_img = loss_img.cuda(args.local_device_rank)
    loss_txt = loss_txt.cuda(args.local_device_rank)

    if sampler is not None:
        sampler.set_epoch(epoch)

    num_steps_per_epoch = dataloader.num_batches // args.accum_freq
    data_iter = iter(dataloader)

    if args.accum_freq > 1:
        accum_images, accum_texts, accum_image_features, accum_text_features = [], [], [], []
        if args.distillation:
            teacher_accum_image_features = []

    end = time.time()
    epoch_trained_steps = 0
    for i in range(0, dataloader.num_batches):
        batch = next(data_iter)

        i_accum = i // args.accum_freq
        step = num_steps_per_epoch * epoch + i_accum
        # reach the args.max_steps, exit training:
        if step >= args.max_steps:
            logging.info("Stopping training due to step {} has reached max_steps {}".format(step, args.max_steps // args.accum_freq))
            return epoch_trained_steps
        scheduler(step)

        optimizer.zero_grad()

        images, texts, eos_indices = batch

        images = images.cuda(args.local_device_rank, non_blocking=True)
        texts = texts.cuda(args.local_device_rank, non_blocking=True)
        eos_indices = eos_indices.cuda(args.local_device_rank, non_blocking=True)

        data_time = time.time() - end

        m = model.module

        if args.accum_freq == 1:
            # with automatic mixed precision.
            if args.precision == "amp":
                with autocast():
                    if args.distillation:
                        total_loss, acc = get_loss(model, images, texts, loss_img, loss_txt, args, teacher_model=teacher_model)
                    else:
                        total_loss, acc = get_loss(model, images, texts, loss_img, loss_txt, args)
                    scaler.scale(total_loss).backward()
                    scaler.step(optimizer)
                scaler.update()

            else:
                if args.distillation:
                    total_loss, acc = get_loss(model, images, texts, loss_img, loss_txt, args, teacher_model=teacher_model)
                else:
                    total_loss, acc = get_loss(model, images, texts, loss_img, loss_txt, args)
                total_loss.backward()
                optimizer.step()
        else:
            # First, cache the features without any gradient tracking.
            with torch.no_grad():
                with autocast(enabled=(args.precision == "amp")):
                    chunk_image_features, chunk_text_features, _ = model(images, texts)
                if args.distillation:
                    output = teacher_model.module.get_feature(images)
                    if(len(output) == 2):
                        teacher_chunk_image_features = output[0]
                    else:
                        teacher_chunk_image_features = output
                accum_image_features.append(chunk_image_features)
                accum_text_features.append(chunk_text_features)
                if args.distillation:
                    teacher_accum_image_features.append(teacher_chunk_image_features)

                accum_images.append(images)
                accum_texts.append(texts)

            # If (i + 1) % accum_freq is not zero, move on to the next batch.
            if ((i + 1) % args.accum_freq) > 0:
                # FIXME this makes data time logging unreliable when accumulating
                continue

            # Now, ready to take gradients for the last accum_freq batches.
            # Re-do the forward pass for those batches, and use the cached features from the other batches as negatives.
            # Call backwards each time, but only step optimizer at the end.
            optimizer.zero_grad()
            for j in range(args.accum_freq):
                images = accum_images[j]
                texts = accum_texts[j]
                with autocast(enabled=(args.precision == "amp")):
                    # `total_loss` and `acc` are coarsely sampled, taking only the last result in the loop.
                    # Although each result should be the same in theory, it will be slightly different in practice
                    if args.distillation:
                        total_loss, acc = get_loss(model, images, texts, loss_img, loss_txt, args, accum_image_features, accum_text_features, j, teacher_model, teacher_accum_image_features)
                    else:
                        total_loss, acc = get_loss(model, images, texts, loss_img, loss_txt, args, accum_image_features, accum_text_features, j)
                if args.precision == "amp":
                    scaler.scale(total_loss).backward()
                else:
                    total_loss.backward()

            if args.precision == "amp":
                scaler.step(optimizer)
                scaler.update()
            else:
                optimizer.step()

        # reset gradient accum, if enabled
        if args.accum_freq > 1:
            accum_images, accum_texts, accum_image_features, accum_text_features = [], [], [], []
            if args.distillation:
                teacher_accum_image_features = []

        # Note: we clamp to 4.6052 = ln(100), as in the original paper.
        m.logit_scale.data = torch.clamp(m.logit_scale.data, 0, 4.6052)

        batch_time = time.time() - end
        end = time.time()

        epoch_trained_steps += 1

        if is_master(args) and ((step + 1) % args.log_interval) == 0:
            batch_size = len(images) * args.accum_freq
            num_samples = (i_accum + 1) * batch_size * args.world_size
            samples_per_epoch = dataloader.num_samples
            percent_complete = 100.0 * (i_accum + 1) / num_steps_per_epoch

            logging.info(
                f"Global Steps: {step + 1}/{args.max_steps} | " +
                f"Train Epoch: {epoch + 1} [{num_samples}/{samples_per_epoch} ({percent_complete:.0f}%)] | " +
                f"Loss: {total_loss.item():.6f} | " +
                (f"Image2Text Acc: {acc['i2t'].item() * 100:.2f} | " if args.report_training_batch_acc else "") +
                (f"Text2Image Acc: {acc['t2i'].item() * 100:.2f} | " if args.report_training_batch_acc else "") +
                f"Data Time: {data_time:.3f}s | " +
                f"Batch Time: {batch_time:.3f}s | " +
                f"LR: {optimizer.param_groups[0]['lr']:5f} | " +
                f"logit_scale: {m.logit_scale.data:.3f} | " +
                f"Global Batch Size: {batch_size * args.world_size}"
            )

        if args.val_data is not None and args.valid_step_interval is not None and ((step + 1) % args.valid_step_interval) == 0:
            assert "val" in data, "Error: Valid dataset has not been built."
            if not args.use_flash_attention:
                evaluate(model, data, epoch, args, step + 1)
            else:
                # fp16 is needed in flash attention
                with autocast():
                    evaluate(model, data, epoch, args, step + 1)
            # set model back to train mode
            model.train()
            if args.freeze_vision:
                freeze_vision_bn(args, model)

        if args.should_save and args.save_step_frequency > 0 and ((step + 1) % args.save_step_frequency) == 0:
            save_path = os.path.join(args.checkpoint_path, f"epoch_{epoch + 1}_{step + 1}.pt")
            t1 = time.time()
            torch.save(
                {
                    "epoch": epoch + 1,
                    "step": step + 1,
                    "name": args.name,
                    "state_dict": model.state_dict() if not args.use_flash_attention else convert_state_dict(model.state_dict()),
                    "optimizer": optimizer.state_dict(),
                },
                save_path,
            )
            logging.info("Saved checkpoint {} (epoch {} @ {} steps) (writing took {} seconds)".format(save_path, epoch + 1, step + 1, time.time() - t1))

            # Save the latest params
            t1 = time.time()
            save_path = os.path.join(args.checkpoint_path, f"epoch_latest.pt")
            torch.save(
                {
                    "epoch": epoch + 1,
                    "step": step + 1,
                    "name": args.name,
                    "state_dict": model.state_dict() if not args.use_flash_attention else convert_state_dict(model.state_dict()),
                    "optimizer": optimizer.state_dict(),
                },
                save_path,
            )
            logging.info("Saved checkpoint {} (epoch {} @ {} steps) (writing took {} seconds)".format(save_path, epoch + 1, step + 1, time.time() - t1))
        
    return epoch_trained_steps


def evaluate(model, data, epoch, args, steps):

    logging.info("Begin to eval on validation set (epoch {} @ {} steps)...".format(epoch + 1, steps))

    model.eval()

    dataloader = data['val'].dataloader
    data_iter = iter(dataloader)

    loss_img = nn.CrossEntropyLoss()
    loss_txt = nn.CrossEntropyLoss()

    loss_img = loss_img.cuda(args.local_device_rank)
    loss_txt = loss_txt.cuda(args.local_device_rank)

    cumulative_loss = torch.zeros([]).cuda(args.local_device_rank, non_blocking=True)
    cumulative_i2t_acc = torch.zeros([]).cuda(args.local_device_rank, non_blocking=True)
    cumulative_t2i_acc = torch.zeros([]).cuda(args.local_device_rank, non_blocking=True)
    num_elements = torch.zeros([]).cuda(args.local_device_rank, non_blocking=True)
    all_image_features, all_text_features = [], []
    with torch.no_grad():
        for i in range(dataloader.num_batches):
            batch = next(data_iter)
            images, texts, eos_indices = batch

            images = images.cuda(args.local_device_rank, non_blocking=True)
            texts = texts.cuda(args.local_device_rank, non_blocking=True)
            eos_indices = eos_indices.cuda(args.local_device_rank, non_blocking=True)

            image_features, text_features, logit_scale = model(images, texts)
            all_image_features.append(image_features)
            all_text_features.append(text_features)
            logit_scale = logit_scale.mean()
            logits_per_image = logit_scale * image_features @ text_features.t()
            logits_per_text = logits_per_image.t()

            ground_truth = torch.arange(len(images)).long()
            ground_truth = ground_truth.cuda(args.local_device_rank, non_blocking=True)
            total_loss = (
                loss_img(logits_per_image, ground_truth)
                + loss_txt(logits_per_text, ground_truth)
            ) / 2

            batch_size = len(images)
            cumulative_loss += total_loss * batch_size
            num_elements += batch_size

            cumulative_i2t_acc += ((logits_per_image.argmax(-1) == ground_truth).sum()).float()
            cumulative_t2i_acc += (logits_per_text.argmax(-1) == ground_truth).sum().float()

            if (i + 1) % 100 == 0:
                logging.info("Evaluated {}/{} batches...".format(i + 1, dataloader.num_batches))

        dist.all_reduce(cumulative_loss, op=dist.ReduceOp.SUM)
        dist.all_reduce(cumulative_i2t_acc, op=dist.ReduceOp.SUM)
        dist.all_reduce(cumulative_t2i_acc, op=dist.ReduceOp.SUM)
        dist.all_reduce(num_elements, op=dist.ReduceOp.SUM)
        loss = cumulative_loss / num_elements
        i2t_acc = cumulative_i2t_acc / num_elements
        t2i_acc = cumulative_t2i_acc / num_elements

        assert num_elements.item() == dataloader.num_samples # sanity check

        logging.info(
            f"Validation Result (epoch {epoch + 1} @ {steps} steps) | "
            f"Valid Loss: {loss.item():.6f} | "
            f"Image2Text Acc: {i2t_acc.item() * 100:.2f} | " 
            f"Text2Image Acc: {t2i_acc.item() * 100:.2f} | " 
            f"logit_scale: {model.module.logit_scale.data:.3f} | "
            f"Valid Batch Size: {batch_size}"
        )

def cosineSimilarityLoss(feature1, feature2):
    scale_factor_h = feature1.shape[0] / feature2.size(0)
    scale_factor_w = feature1.shape[1] / feature2.size(1)

    feature2_interpolated = F.interpolate(feature2.unsqueeze(0).unsqueeze(0),
                            size=(feature1.shape[0], feature1.shape[1]),
                            mode='bilinear',
                            align_corners=False)
    feature2_interpolated = feature2_interpolated.squeeze(0).squeeze(0)
    

    cosine_sim = F.cosine_similarity(feature1, feature2_interpolated, dim=1)
    similarity_loss = 1 - cosine_sim.mean()
    return similarity_loss


import os
import logging
import json
from dataclasses import dataclass
from pathlib import Path
from PIL import Image
import base64
from io import BytesIO
import torch
import lmdb
from torchvision.transforms import Compose, Resize, ToTensor, Normalize, InterpolationMode
from torch.utils.data import Dataset, DataLoader, SubsetRandomSampler
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data.sampler import SequentialSampler
import torchvision.datasets as datasets
from cn_clip.clip import tokenize


def _convert_to_rgb(image):
    return image.convert('RGB')


def _preprocess_text(text):
    # adapt the text to Chinese BERT vocab
    text = text.lower().replace("“", "\"").replace("”", "\"")
    return text


class EvalTxtDataset(Dataset):
    def __init__(self, jsonl_filename, max_txt_length=24):
        assert os.path.exists(jsonl_filename), "The annotation datafile {} not exists!".format(jsonl_filename)

        logging.debug(f'Loading jsonl data from {jsonl_filename}.')
        self.texts = []
        with open(jsonl_filename, "r", encoding="utf-8") as fin:
            for line in fin:
                obj = json.loads(line.strip())
                text_id = obj['text_id']
                text = obj['text']
                self.texts.append((text_id, text))
        logging.debug(f'Finished loading jsonl data from {jsonl_filename}.')

        self.max_txt_length = max_txt_length

    def __len__(self):
        return len(self.texts)

    def __getitem__(self, idx):
        text_id, text = self.texts[idx]
        text = tokenize([_preprocess_text(str(text))], context_length=self.max_txt_length)[0]
        return text_id, text


class EvalImgDataset(Dataset):
    def __init__(self, lmdb_imgs, resolution=224):
        assert os.path.isdir(lmdb_imgs), "The image LMDB directory {} not exists!".format(lmdb_imgs)

        logging.debug(f'Loading image LMDB from {lmdb_imgs}.')

        self.env_imgs = lmdb.open(lmdb_imgs, readonly=True, create=False, lock=False, readahead=False, meminit=False)
        self.txn_imgs = self.env_imgs.begin(buffers=True)
        self.cursor_imgs = self.txn_imgs.cursor()
        self.iter_imgs = iter(self.cursor_imgs)
        self.number_images = int(self.txn_imgs.get(key=b'num_images').tobytes().decode('utf-8'))
        logging.info("The specified LMDB directory contains {} images.".format(self.number_images))

        self.transform = self._build_transform(resolution)

    def _build_transform(self, resolution):
        normalize = Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711))
        return Compose([
                Resize((resolution, resolution), interpolation=InterpolationMode.BICUBIC),
                _convert_to_rgb,
                ToTensor(),
                normalize,
            ])

    def __len__(self):
        return self.number_images

    def __getitem__(self, idx):
        img_id, image_b64 = next(self.iter_imgs)
        if img_id == b"num_images":
            img_id, image_b64 = next(self.iter_imgs)

        img_id = img_id.tobytes()
        image_b64 = image_b64.tobytes()

        img_id = int(img_id.decode(encoding="utf8", errors="ignore"))
        image_b64 = image_b64.decode(encoding="utf8", errors="ignore")
        image = Image.open(BytesIO(base64.urlsafe_b64decode(image_b64))) # already resized
        image = self.transform(image)

        return img_id, image


@dataclass
class DataInfo:
    dataloader: DataLoader
    sampler: DistributedSampler


def get_eval_txt_dataset(args, max_txt_length=24):
    input_filename = args.text_data
    dataset = EvalTxtDataset(
        input_filename,
        max_txt_length=max_txt_length)
    num_samples = len(dataset)
    sampler = SequentialSampler(dataset)

    dataloader = DataLoader(
        dataset,
        batch_size=args.text_batch_size,
        num_workers=0,
        pin_memory=True,
        sampler=sampler,
        drop_last=False,
    )
    dataloader.num_samples = num_samples
    dataloader.num_batches = len(dataloader)

    return DataInfo(dataloader, sampler)


def fetch_resolution(vision_model):
    # fetch the resolution from the vision model config
    vision_model_config_file = Path(__file__).parent.parent / f"clip/model_configs/{vision_model.replace('/', '-')}.json"
    with open(vision_model_config_file, 'r') as fv:
        model_info = json.load(fv)
    return model_info["image_resolution"]


def get_eval_img_dataset(args):
    lmdb_imgs = args.image_data
    dataset = EvalImgDataset(
        lmdb_imgs, resolution=fetch_resolution(args.vision_model))
    num_samples = len(dataset)
    sampler = SequentialSampler(dataset)

    dataloader = DataLoader(
        dataset,
        batch_size=args.img_batch_size,
        num_workers=0,
        pin_memory=True,
        sampler=sampler,
        drop_last=False,
    )
    dataloader.num_samples = num_samples
    dataloader.num_batches = len(dataloader)

    return DataInfo(dataloader, sampler)


def get_zeroshot_dataset(args, preprocess_fn):
    dataset = datasets.ImageFolder(args.datapath, transform=preprocess_fn)

    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=args.img_batch_size,
        num_workers=args.num_workers,
        sampler=None,
    )

    return DataInfo(dataloader, None)


from contextlib import contextmanager
import os
from time import time
import dataclasses
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional
import torch

import tensorrt as trt
from tensorrt import ICudaEngine, ILayer, INetworkDefinition, Logger, Runtime
from tensorrt.tensorrt import Builder, IBuilderConfig, IElementWiseLayer, IOptimizationProfile, IReduceLayer


@dataclass
class TensorRTShape:
    """
    Store input shapes for TensorRT build.
    3 shapes per input tensor are required (as tuple of integers):

    * minimum input shape
    * optimal size used for the benchmarks during building
    * maximum input shape

    Set input name to None for default shape.
    """

    min_shape: List[int]
    optimal_shape: List[int]
    max_shape: List[int]
    input_name: Optional[str]

    def check_validity(self) -> None:
        """
        Basic checks of provided shapes
        """
        assert len(self.min_shape) == len(self.optimal_shape) == len(self.max_shape)
        assert len(self.min_shape) > 0
        assert self.min_shape[0] > 0 and self.optimal_shape[0] > 0 and self.max_shape[0] > 0
        assert self.input_name is not None

    def make_copy(self, input_name: str) -> "TensorRTShape":
        """
        Make a copy of the current instance, with a different input name.
        :param input_name: new input name to use
        :return: a copy of the current shape with a different name
        """
        instance_copy = dataclasses.replace(self)
        instance_copy.input_name = input_name
        return instance_copy

    def generate_multiple_shapes(self, input_names: List[str]) -> List["TensorRTShape"]:
        """
        Generate multiple shapes when only a single default one is defined.
        :param input_names: input names used by the model
        :return: a list of shapes
        """
        assert self.input_name is None, f"input name is not None: {self.input_name}"
        result = list()
        for name in input_names:
            shape = self.make_copy(input_name=name)
            result.append(shape)
        return result


def fix_fp16_network(network_definition: INetworkDefinition,
                     fp16_banned_ops = None) -> INetworkDefinition:
    """
    Mixed precision on TensorRT can generate scores very far from Pytorch because of some operator being saturated.
    Indeed, FP16 can't store very large and very small numbers like FP32.
    Here, we search for some patterns of operators to keep in FP32, in most cases, it is enough to fix the inference
    and don't hurt performances.
    :param network_definition: graph generated by TensorRT after parsing ONNX file (during the model building)
    :return: patched network definition
    """
    # search for patterns which may overflow in FP16 precision, we force FP32 precisions for those nodes
    for layer_index in range(network_definition.num_layers - 1):
        layer: ILayer = network_definition.get_layer(layer_index)
        next_layer: ILayer = network_definition.get_layer(layer_index + 1)
        print("layer name:{}; type:{}".format(layer.name, layer.type))

        if fp16_banned_ops is not None:
            for cand in fp16_banned_ops:
                if layer.name.find(cand) != -1:
                    layer.precision = trt.DataType.FLOAT
                    layer.set_output_type(index=0, dtype=trt.DataType.FLOAT)

        # POW operation usually followed by mean reduce(transformer-deploy's strategy)
        if layer.type == trt.LayerType.ELEMENTWISE and next_layer.type == trt.LayerType.REDUCE:
            # casting to get access to op attribute
            layer.__class__ = IElementWiseLayer
            next_layer.__class__ = IReduceLayer
            if layer.op == trt.ElementWiseOperation.POW:
                layer.precision = trt.DataType.FLOAT
                next_layer.precision = trt.DataType.FLOAT
            layer.set_output_type(index=0, dtype=trt.DataType.FLOAT)
            next_layer.set_output_type(index=0, dtype=trt.DataType.FLOAT)
    return network_definition


def build_engine(
    runtime: Runtime,
    onnx_file_path: str,
    logger: Logger,
    fp16: bool,
    int8: bool,
    workspace_size: Optional[int] = None,
    fp16_fix: Callable[[INetworkDefinition], INetworkDefinition] = fix_fp16_network,
    fp16_banned_ops = None,
    timing_cache=None,
    **kwargs,
) -> ICudaEngine:
    """
    Convert ONNX file to TensorRT engine.
    It supports dynamic shape, however it's advised to keep sequence length fix as it hurts performance otherwise.
    Dynamic batch size doesn't hurt performance and is highly advised.
    Batch size can provided through different ways:
    * **min_shape**, **optimal_shape**, **max_shape**: for simple case, 3 tuples of int when all
    input tensors have the same shape
    * **input_shapes**: a list of TensorRTShape with names if there are several input tensors with different shapes
    **TIP**: minimum batch size should be 1 in most cases.
    :param runtime: global variable shared accross inference call / model building
    :param onnx_file_path: path to the ONNX file
    :param logger: specific logger to TensorRT
    :param workspace_size: GPU memory to use during the building, more is always better.
        If there is not enough memory, some optimization may fail, and the whole conversion process will crash.
    :param fp16: enable FP16 precision, it usually provide a 20-30% boost compared to ONNX Runtime.
    :param int8: enable INT-8 quantization, best performance but model should have been quantized.
    :param fp16_fix: a function to set FP32 precision on some nodes to fix FP16 overflow
    :return: TensorRT engine to use during inference
    """
    # default input shape
    if "min_shape" in kwargs and "optimal_shape" in kwargs and "max_shape" in kwargs:
        default_shape = TensorRTShape(
            min_shape=kwargs["min_shape"],
            optimal_shape=kwargs["optimal_shape"],
            max_shape=kwargs["max_shape"],
            input_name=None,
        )
        input_shapes = [default_shape]
    else:
        assert "input_shapes" in kwargs, "missing input shapes"
        input_shapes: List[TensorRTShape] = kwargs["input_shapes"]

    builder: Builder = trt.Builder(logger)
    config: IBuilderConfig = builder.create_builder_config()
    if workspace_size is not None:
        config.set_memory_pool_limit(trt.tensorrt.MemoryPoolType.DLA_GLOBAL_DRAM, workspace_size)
    config.set_tactic_sources(
        tactic_sources=1 << int(trt.TacticSource.CUBLAS)
        | 1 << int(trt.TacticSource.CUBLAS_LT)
        | 1 << int(trt.TacticSource.CUDNN)  # trt advised to use cuDNN for transfo architecture
    )
    if int8:
        config.set_flag(trt.BuilderFlag.INT8)
    if fp16:
        config.set_flag(trt.BuilderFlag.FP16)

    # config.set_flag(trt.BuilderFlag.DISABLE_TIMING_CACHE)
    timing_cache_available = int(trt.__version__[0]) >= 8 and timing_cache != None
    # load global timing cache
    if timing_cache_available:
        if os.path.exists(timing_cache):
            with open(timing_cache, "rb") as f:
                cache = config.create_timing_cache(f.read())
                config.set_timing_cache(cache, ignore_mismatch=False)
        else:
            cache = config.create_timing_cache(b"")
            config.set_timing_cache(cache, ignore_mismatch=False)

    explicit_batch = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
    network_def = builder.create_network(explicit_batch)
    # https://github.com/NVIDIA/TensorRT/issues/1196 (sometimes big diff in output when using FP16)
    config.set_flag(trt.BuilderFlag.OBEY_PRECISION_CONSTRAINTS)
    logger.log(msg="parsing TensorRT model", severity=trt.ILogger.INFO)

    with trt.OnnxParser(network_def, logger) as parser:
        with open(onnx_file_path, "rb") as f:
            # file path needed for models with external dataformat
            # https://github.com/onnx/onnx-tensorrt/issues/818
            parsed = parser.parse(f.read(), onnx_file_path)
            for i in range(parser.num_errors):
                print("TensorRT ONNX parser error:", parser.get_error(i))

    profile: IOptimizationProfile = builder.create_optimization_profile()
    # duplicate default shape (one for each input)
    if len(input_shapes) == 1 and input_shapes[0].input_name is None:
        names = [network_def.get_input(num_input).name for num_input in range(network_def.num_inputs)]
        input_shapes = input_shapes[0].generate_multiple_shapes(input_names=names)

    for shape in input_shapes:
        shape.check_validity()
        profile.set_shape(
            input=shape.input_name,
            min=shape.min_shape,
            opt=shape.optimal_shape,
            max=shape.max_shape,
        )
    if "shape_tensors" in kwargs:
        for shape in kwargs["shape_tensors"]:
            profile.set_shape_input(
                input=shape.input_name,
                min=shape.min_shape,
                opt=shape.optimal_shape,
                max=shape.max_shape,
            )
    config.add_optimization_profile(profile)
    if fp16:
        network_def = fp16_fix(network_def, fp16_banned_ops)

    logger.log(msg="building engine. depending on model size this may take a while", severity=trt.ILogger.WARNING)
    start = time()
    # trt_engine = builder.build_serialized_network(network_def, config)
    # engine: ICudaEngine = runtime.deserialize_cuda_engine(trt_engine)
    engine = builder.build_engine(network_def, config=config)
    logger.log(msg=f"building engine took {time() - start:4.1f} seconds", severity=trt.ILogger.WARNING)
    assert engine is not None, "error during engine generation, check error messages above :-("
    # save global timing cache
    if timing_cache_available:
        cache = config.get_timing_cache()
        with cache.serialize() as buffer:
            with open(timing_cache, "wb") as f:
                f.write(buffer)
                f.flush()
                os.fsync(f)
    return engine


@contextmanager
def track_infer_time(buffer: List[int]) -> None:
    """
    A context manager to perform latency measures
    :param buffer: a List where to save latencies for each input
    """
    start = time.perf_counter()
    yield
    end = time.perf_counter()
    buffer.append(end - start)


TRT_LOGGER = trt.Logger()


def get_binding_idxs(engine: trt.ICudaEngine, profile_index: int):
    """
    Calculate start/end binding indices for current context's profile
    https://docs.nvidia.com/deeplearning/tensorrt/developer-guide/index.html#opt_profiles_bindings
    :param engine: TensorRT engine generated during the model building
    :param profile_index: profile to use (several profiles can be set during building)
    :return: input and output tensor indexes
    """
    num_bindings_per_profile = engine.num_bindings // engine.num_optimization_profiles
    start_binding = profile_index * num_bindings_per_profile
    end_binding = start_binding + num_bindings_per_profile  # Separate input and output binding indices for convenience
    input_binding_idxs: List[int] = []
    output_binding_idxs: List[int] = []
    for binding_index in range(start_binding, end_binding):
        if engine.binding_is_input(binding_index):
            input_binding_idxs.append(binding_index)
        else:
            output_binding_idxs.append(binding_index)
    return input_binding_idxs, output_binding_idxs


def get_output_tensors(
    context: trt.IExecutionContext,
    host_inputs: List[torch.Tensor],
    input_binding_idxs: List[int],
    output_binding_idxs: List[int],
):
    """
    Reserve memory in GPU for input and output tensors.
    :param context: TensorRT context shared accross inference steps
    :param host_inputs: input tensor
    :param input_binding_idxs: indexes of each input vector (should be the same than during building)
    :param output_binding_idxs: indexes of each output vector (should be the same than during building)
    :return: tensors where output will be stored
    """
    # explicitly set dynamic input shapes, so dynamic output shapes can be computed internally
    for host_input, binding_index in zip(host_inputs, input_binding_idxs):
        input_name = context.engine.get_binding_name(binding_index)
        context.set_binding_shape(binding_index, tuple(host_input.shape))
    # assert context.all_binding_shapes_specified
    device_outputs: Dict[str, torch.Tensor] = dict()
    for binding_index in output_binding_idxs:
        # TensorRT computes output shape based on input shape provided above
        output_shape = context.get_binding_shape(binding=binding_index)
        output_name = context.engine.get_binding_name(index=binding_index)
        # allocate buffers to hold output results
        device_outputs[output_name] = torch.empty(tuple(output_shape), device="cuda")
    return device_outputs


class TensorRTModel(object):
    def __init__(self, engine_path):
        print(f'load engine_path is {engine_path}')
        self.engine = self.load_engine(engine_path)
        profile_index = 0
        self.context = self.engine.create_execution_context()
        self.context.set_optimization_profile_async(
            profile_index=profile_index, stream_handle=torch.cuda.current_stream().cuda_stream
        )
        self.input_binding_idxs, self.output_binding_idxs = get_binding_idxs(self.engine, profile_index)

    def load_engine(self, engine_file_path):
        assert os.path.exists(engine_file_path)
        print("Reading engine from file {}".format(engine_file_path))
        with open(engine_file_path, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())

    def __call__(self, inputs, time_buffer=None):
        input_tensors: List[torch.Tensor] = list()
        for i in range(self.context.engine.num_bindings):
            if not self.context.engine.binding_is_input(index=i):
                continue
            tensor_name = self.context.engine.get_binding_name(i)
            assert tensor_name in inputs, f"input not provided: {tensor_name}"
            tensor = inputs[tensor_name]
            assert isinstance(tensor, torch.Tensor), f"unexpected tensor class: {type(tensor)}"
            assert tensor.device.type == "cuda", f"unexpected device type (trt only works on CUDA): {tensor.device.type}"
            # warning: small changes in output if int64 is used instead of int32
            if tensor.dtype in [torch.int64, torch.long]:
                # logging.warning(f"using {tensor.dtype} instead of int32 for {tensor_name}, will be casted to int32")
                tensor = tensor.type(torch.int32)
            input_tensors.append(tensor)

        # calculate input shape, bind it, allocate GPU memory for the output
        outputs: Dict[str, torch.Tensor] = get_output_tensors(
            self.context, input_tensors, self.input_binding_idxs, self.output_binding_idxs
        )
        bindings = [int(i.data_ptr()) for i in input_tensors + list(outputs.values())]
        if time_buffer is None:
            self.context.execute_v2(bindings=bindings)
        else:
            with track_infer_time(time_buffer):
                self.context.execute_v2(bindings=bindings)

        torch.cuda.current_stream().synchronize()  # sync all CUDA ops

        return outputs


# -*- coding: utf-8 -*-
'''
This script extracts image and text features for evaluation using ONNX model. (with CPU or single-GPU)
'''
import os
import sys
import argparse
import json
import torch
from tqdm import tqdm
import onnxruntime
from cn_clip.eval.data import get_eval_img_dataset, get_eval_txt_dataset

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--extract-image-feats', 
        action="store_true", 
        default=False, 
        help="Whether to extract image features."
    )
    parser.add_argument(
        '--extract-text-feats', 
        action="store_true", 
        default=False, 
        help="Whether to extract text features."
    )
    parser.add_argument(
        '--image-data', 
        type=str, 
        default="../Multimodal_Retrieval/lmdb/test/imgs", 
        help="If --extract-image-feats is True, specify the path of the LMDB directory storing input image base64 strings."
    )
    parser.add_argument(
        '--text-data', 
        type=str, 
        default="../Multimodal_Retrieval/test_texts.jsonl", 
        help="If --extract-text-feats is True, specify the path of input text Jsonl file."
    )
    parser.add_argument(
        '--image-feat-output-path', 
        type=str, 
        default=None, 
        help="If --extract-image-feats is True, specify the path of output image features."
    )    
    parser.add_argument(
        '--text-feat-output-path', 
        type=str, 
        default=None, 
        help="If --extract-image-feats is True, specify the path of output text features."
    )
    parser.add_argument(
        "--img-batch-size", type=int, default=1, help="Image batch size."
    )
    parser.add_argument(
        "--text-batch-size", type=int, default=1, help="Text batch size."
    )    
    parser.add_argument(
        "--context-length", type=int, default=52, help="The maximum length of input text (include [CLS] & [SEP] tokens)."
    )    
    parser.add_argument(
        "--onnx-image-model",
        default=None,
        type=str,
        help="Path to ONNX image model.",
    )
    parser.add_argument(
        "--onnx-text-model",
        default=None,
        type=str,
        help="Path to ONNX text model.",
    )
    parser.add_argument(
        "--device",
        choices=["cuda", "cpu"],
        default="cuda",
        help="Use CPU or GPU ONNX runtime.",
    )
    parser.add_argument(
        "--vision-model",
        choices=["ViT-B-16", "ViT-L-14", "ViT-L-14-336", "ViT-H-14", "RN50"],
        default="ViT-B-16",
        help="Name of the vision backbone to use.",
    )
    parser.add_argument(
        "--text-model",
        choices=["RoBERTa-wwm-ext-base-chinese", "RoBERTa-wwm-ext-large-chinese", "RBT3-chinese"],
        default="RoBERTa-wwm-ext-base-chinese",
        help="Name of the text backbone to use.",
    )
    parser.add_argument(
        "--debug",
        default=False,
        action="store_true",
        help="If true, more information is logged."
    )
    args = parser.parse_args()

    return args


if __name__ == "__main__":
    args = parse_args()

    assert args.extract_image_feats or args.extract_text_feats, "--extract-image-feats and --extract-text-feats cannot both be False!"

    # Log params.
    print("Params:")
    for name in sorted(vars(args)):
        val = getattr(args, name)
        print(f"  {name}: {val}")

    # Get data.
    if args.extract_image_feats:
        print("Preparing image inference dataset.")
        img_data = get_eval_img_dataset(args)
    if args.extract_text_feats:
        print("Preparing text inference dataset.")
        text_data = get_eval_txt_dataset(args, max_txt_length=args.context_length)

    if args.device == "cpu":
        provider = "CPUExecutionProvider" 
    else:
        provider = "CUDAExecutionProvider"

    img_sess_options = onnxruntime.SessionOptions()
    img_run_options = onnxruntime.RunOptions()
    img_run_options.log_severity_level = 2
    img_session = onnxruntime.InferenceSession(args.onnx_image_model,
                                               sess_options=img_sess_options,
                                               providers=[provider])

    txt_sess_options = onnxruntime.SessionOptions()
    txt_run_options = onnxruntime.RunOptions()
    txt_run_options.log_severity_level = 2
    txt_session = onnxruntime.InferenceSession(args.onnx_text_model,
                                               sess_options=txt_sess_options,
                                               providers=[provider])

    # Make inference for texts
    if args.extract_text_feats:
        print('Make inference for texts...')
        if args.text_feat_output_path is None:
            args.text_feat_output_path = "{}.txt_feat.jsonl".format(args.text_data[:-6])
        write_cnt = 0
        with open(args.text_feat_output_path, "w") as fout:
            dataloader = text_data.dataloader
            with torch.no_grad():
                for batch in tqdm(dataloader):
                    text_ids, texts = batch
                    text_features = txt_session.run(["unnorm_text_features"], {"text": texts.numpy()})[0]
                    text_features = torch.tensor(text_features).to(args.device)
                    text_features /= text_features.norm(dim=-1, keepdim=True)
                    for text_id, text_feature in zip(text_ids.tolist(), text_features.tolist()):
                        fout.write("{}\n".format(json.dumps({"text_id": text_id, "feature": text_feature})))
                        write_cnt += 1
        print('{} text features are stored in {}'.format(write_cnt, args.text_feat_output_path))

    # Make inference for images
    if args.extract_image_feats:
        print('Make inference for images...')
        if args.image_feat_output_path is None:
            # by default, we store the image features under the same directory with the text features
            args.image_feat_output_path = "{}.img_feat.jsonl".format(
                args.text_data.replace("_texts.jsonl", "_imgs"))
        write_cnt = 0
        with open(args.image_feat_output_path, "w") as fout:
            dataloader = img_data.dataloader
            with torch.no_grad():
                for batch in tqdm(dataloader):
                    image_ids, images = batch
                    image_features = img_session.run(["unnorm_image_features"], {"image": images.numpy()})[0]
                    image_features = torch.tensor(image_features).to(args.device)
                    image_features /= image_features.norm(dim=-1, keepdim=True)
                    for image_id, image_feature in zip(image_ids.tolist(), image_features.tolist()):
                        fout.write("{}\n".format(json.dumps({"image_id": image_id, "feature": image_feature})))
                        write_cnt += 1
        print('{} image features are stored in {}'.format(write_cnt, args.image_feat_output_path))

    print("Done!")


# -*- coding: utf-8 -*-
'''
This script performs zero-shot evaluation on ImageNet-1K. (with single-GPU)
'''

import os
import argparse
from pathlib import Path
import json
from tqdm import tqdm

import torch

from cn_clip.clip.model import convert_weights, CLIP
from cn_clip.clip import tokenize
from cn_clip.training.main import convert_models_to_fp32
from cn_clip.clip.utils import image_transform
from cn_clip.eval.data import get_zeroshot_dataset, _preprocess_text
from cn_clip.eval.cvinw_zeroshot_templates import (
    openai_templates,
    flower_templates,
    food_templates,
    aircraft_templates,
    eurosat_templates,
    country211_templates,
)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--vision-model",
        choices=["ViT-B-32", "ViT-B-16", "ViT-L-14", "ViT-L-14-336", "ViT-H-14", "RN50"],
        default="ViT-B-16",
        help="Name of the vision backbone to use.",
    )
    parser.add_argument(
        "--text-model",
        choices=["RoBERTa-wwm-ext-base-chinese", "RoBERTa-wwm-ext-large-chinese", "RBT3-chinese"],
        default="RoBERTa-wwm-ext-base-chinese",
        help="Name of the text backbone to use.",
    )
    parser.add_argument(
        "--precision",
        choices=["amp", "fp16", "fp32"],
        default="amp",
        help="Floating point precition."
    )
    parser.add_argument(
        "--label-file",
        type=str,
        help="file for labels",
    )
    parser.add_argument(
        "--datapath",
        type=str,
        required=True,
        help="Path to the test set for conducting zero shot evaluation.",
    )
    parser.add_argument(
        "--dataset",
        type=str,
        default="imagenet",
        help="Specified dataset.",
    )
    parser.add_argument(
        "--index",
        type=str,
        default="",
        help="Specify image paths.",
    )
    parser.add_argument(
        "--save-dir",
        type=str,
        default="",
        help="Specified dataset.",
    )
    # parser.add_argument(
    #     "--imagenet-val",
    #     type=str,
    #     required=True,
    #     help="Path to imagenet val set for conducting zero shot evaluation.",
    # )
    parser.add_argument(
        "--img-batch-size", type=int, default=64, help="Image batch size."
    )    
    parser.add_argument(
        "--context-length", 
        type=int, 
        default=52,
        help="The maximum length of input text (include [CLS] & [SEP] tokens)."
    )
    parser.add_argument(
        "--resume",
        default=None,
        type=str,
        help="path to latest checkpoint (default: none)",
    )    
    parser.add_argument(
        "--num-workers", type=int, default=4, help="Number of workers for ImageNet dataloader."
    )        
    args = parser.parse_args()

    return args


def zero_shot_classifier(model, classnames, templates, args):
    with torch.no_grad():
        zeroshot_weights = []
        for classname in tqdm(classnames):
            texts = [_preprocess_text(template(classname)) for template in templates]  # format with class
            texts = tokenize(texts, context_length=args.context_length).to(args.gpu)  # tokenize
            class_embeddings = model(None, texts)
            class_embeddings /= class_embeddings.norm(dim=-1, keepdim=True)
            class_embedding = class_embeddings.mean(dim=0)
            class_embedding /= class_embedding.norm()
            zeroshot_weights.append(class_embedding)
        zeroshot_weights = torch.stack(zeroshot_weights, dim=1).to(args.gpu)
    return zeroshot_weights


def accuracy(output, target, topk=(1,)):
    pred = output.topk(max(topk), 1, True, True)[1].t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    return [float(correct[:k].reshape(-1).float().sum(0, keepdim=True).cpu().numpy()) for k in topk]


def run(model, classifier, dataloader, args):
    total_logits = []
    total_targets = []
    with torch.no_grad():
        top1, top5, n = 0.0, 0.0, 0.0
        for images, target in tqdm(dataloader):
            images = images.to(args.gpu)
            target = target.to(args.gpu)
            total_targets.append(target)

            # predict
            image_features = model(images, None)
            image_features /= image_features.norm(dim=-1, keepdim=True)
            logits = (100.0 * image_features @ classifier).softmax(dim=-1)
            total_logits.append(logits)

            # measure accuracy
            acc1, acc5 = accuracy(logits, target, topk=(1, 1))
            top1 += acc1
            n += images.size(0)

    outputs = torch.cat(total_logits, dim=0)
    targets = torch.cat(total_targets, dim=0)

    if getattr(args, "index", ""):
        print("Use index to rearrange the logits...")
        with open(args.index, "r", encoding="utf-8") as f:
            index = json.load(f)
            print(index)
        outputs = outputs[index]
        targets = targets[index]
        print(targets)

    top1 = top1 / n

    return top1, outputs


if __name__ == "__main__":
    args = parse_args()

    # Log params.
    print("Params:")
    for name in sorted(vars(args)):
        val = getattr(args, name)
        print(f"  {name}: {val}")

    args.gpu = 0
    torch.cuda.set_device(args.gpu)

    # Initialize the model.
    vision_model_config_file = Path(__file__).parent.parent / f"clip/model_configs/{args.vision_model.replace('/', '-')}.json"
    print('Loading vision model config from', vision_model_config_file)
    assert os.path.exists(vision_model_config_file)
    
    text_model_config_file = Path(__file__).parent.parent / f"clip/model_configs/{args.text_model.replace('/', '-')}.json"
    print('Loading text model config from', text_model_config_file)
    assert os.path.exists(text_model_config_file)
    
    with open(vision_model_config_file, 'r') as fv, open(text_model_config_file, 'r') as ft:
        model_info = json.load(fv)
        if isinstance(model_info['vision_layers'], str):
            model_info['vision_layers'] = eval(model_info['vision_layers'])        
        for k, v in json.load(ft).items():
            model_info[k] = v

    model = CLIP(**model_info)
    convert_weights(model)

    # See https://discuss.pytorch.org/t/valueerror-attemting-to-unscale-fp16-gradients/81372
    if args.precision == "amp" or args.precision == "fp32":
        convert_models_to_fp32(model)
    model.cuda(args.gpu)
    if args.precision == "fp16":
        convert_weights(model)

    # Get eval data.
    print("Preparing zeroshot dataset.")
    data = {}
    print(f"{model_info['image_resolution']}")
    data[args.dataset] = get_zeroshot_dataset(
        args, image_transform(model_info["image_resolution"])
    )

    # Resume from a checkpoint.
    print("Begin to load model checkpoint from {}.".format(args.resume))
    assert os.path.exists(args.resume), "The checkpoint file {} not exists!".format(args.resume)
    # Map model to be loaded to specified single gpu.
    loc = "cuda:{}".format(args.gpu)
    checkpoint = torch.load(args.resume, map_location='cpu')
    start_epoch = checkpoint["epoch"]
    sd = checkpoint["state_dict"]
    if next(iter(sd.items()))[0].startswith('module'):
        sd = {k[len('module.'):]: v for k, v in sd.items() if "bert.pooler" not in k}
    model.load_state_dict(sd)
    print(
        f"=> loaded checkpoint '{args.resume}' (epoch {checkpoint['epoch']} @ {checkpoint['step']} steps)"
    )

    # Compute ensembled class embeddings
    print('Building zero-shot classifier')

    model.eval()

    f = open(args.label_file, "r", encoding="utf8")
    classnames = [line.strip() for line in f.readlines()]

    template_dict = {
        "fgvc-aircraft-2013b-variants102": aircraft_templates,
        "food-101": food_templates,
        "oxford-flower-102": flower_templates,
        "eurosat_clip": eurosat_templates,
        "resisc45_clip": eurosat_templates,
        "country211": country211_templates,
        "openai": openai_templates,
    }
    if args.dataset in template_dict.keys():
        templates = template_dict[args.dataset]
    else:
        templates = template_dict['openai']

    # Make inference and evaluation
    print('Using classifier')
    classifier = zero_shot_classifier(model, classnames, templates, args)
    results = {}
    top1, logits = run(model, classifier, data[args.dataset].dataloader, args)

    def json_prec_dump(data, prec=6):
        return json.dumps(
            json.loads(json.dumps(data), parse_float=lambda x: round(float(x), prec))
        )

    print(logits.size())
    output_dict = {
        "model_name": "CN-CLIP-" + args.vision_model,
        "dataset_name": args.dataset,
        "num_trainable_params": 0,
        "num_params": sum(x.numel() for x in model.parameters()),
        "num_visual_params": sum(x.numel() for x in model.visual.parameters()),
        "num_backbone_params": sum(x.numel() for x in model.parameters()),
        "n_shot": 0,
        "rnd_seeds": [123],
        "predictions": [logits.cpu().data.numpy().tolist()],
    }
    json_string = json_prec_dump(output_dict)
    with open(os.path.join(args.save_dir, f"{args.dataset}.json"), "w", encoding="utf-8") as w:
        w.write(json_string)

    results["zeroshot-top1"] = top1

    print('Result:')
    print(", ".join(["{}: {}".format(k, v) for k, v in results.items()]))
    print('Finished.')



"""
This script provides templates for manual prompting for zero-shot image classification.
"""


openai_templates = [
    lambda c: f"{c}的照片",
    lambda c: f"质量差的{c}的照片",
    lambda c: f"许多{c}的照片",
    lambda c: f"{c}的雕塑",
    lambda c: f"难以看到{c}的照片",
    lambda c: f"{c}的低分辨率照片",
    lambda c: f"{c}的渲染",
    lambda c: f"涂鸦{c}",
    lambda c: f"{c}的糟糕照片",
    lambda c: f"{c}的裁剪照片",
    lambda c: f"{c}的纹身",
    lambda c: f"{c}的刺绣照片",
    lambda c: f"很难看到{c}的照片",
    lambda c: f"{c}的明亮照片",
    lambda c: f"一张干净的{c}的照片",
    lambda c: f"一张包含{c}的照片",
    lambda c: f"{c}的深色照片",
    lambda c: f"{c}的手绘画",
    lambda c: f"我的{c}的照片",
    lambda c: f"不自然的{c}的照片",
    lambda c: f"一张酷的{c}的照片",
    lambda c: f"{c}的特写照片",
    lambda c: f"{c}的黑白照片",
    lambda c: f"一幅{c}的画",
    lambda c: f"一幅{c}的绘画",
    lambda c: f"一张{c}的像素照片",
    lambda c: f"{c}的雕像",
    lambda c: f"一张{c}的明亮照片",
    lambda c: f"{c}的裁剪照片",
    lambda c: f"人造的{c}的照片",
    lambda c: f"一张关于{c}的照片",
    lambda c: f"损坏的{c}的jpeg照片",
    lambda c: f"{c}的模糊照片",
    lambda c: f"{c}的相片",
    lambda c: f"一张{c}的好照片",
    lambda c: f"{c}的渲染照",
    lambda c: f"视频游戏中的{c}",
    lambda c: f"一张{c}的照片",
    lambda c: f"{c}的涂鸦",
    lambda c: f"{c}的近距离照片",
    lambda c: f"{c}的折纸",
    lambda c: f"{c}在视频游戏中",
    lambda c: f"{c}的草图",
    lambda c: f"{c}的涂鸦照",
    lambda c: f"{c}的折纸形状",
    lambda c: f"低分辨率的{c}的照片",
    lambda c: f"玩具{c}",
    lambda c: f"{c}的副本",
    lambda c: f"{c}的干净的照片",
    lambda c: f"一张大{c}的照片",
    lambda c: f"{c}的重现",
    lambda c: f"一张漂亮的{c}的照片",
    lambda c: f"一张奇怪的{c}的照片",
    lambda c: f"模糊的{c}的照片",
    lambda c: f"卡通{c}",
    lambda c: f"{c}的艺术作品",
    lambda c: f"{c}的素描",
    lambda c: f"刺绣{c}",
    lambda c: f"{c}的像素照",
    lambda c: f"{c}的拍照",
    lambda c: f"{c}的损坏的照片",
    lambda c: f"高质量的{c}的照片",
    lambda c: f"毛绒玩具{c}",
    lambda c: f"漂亮的{c}的照片",
    lambda c: f"小{c}的照片",
    lambda c: f"照片是奇怪的{c}",
    lambda c: f"漫画{c}",
    lambda c: f"{c}的艺术照",
    lambda c: f"{c}的图形",
    lambda c: f"大{c}的照片",
    lambda c: f"黑白的{c}的照片",
    lambda c: f"{c}毛绒玩具",
    lambda c: f"一张{c}的深色照片",
    lambda c: f"{c}的摄影图",
    lambda c: f"{c}的涂鸦照",
    lambda c: f"玩具形状的{c}",
    lambda c: f"拍了{c}的照片",
    lambda c: f"酷酷的{c}的照片",
    lambda c: f"照片里的小{c}",
    lambda c: f"{c}的刺青",
    lambda c: f"{c}的可爱的照片",
    lambda c: f"一张{c}可爱的照片",
    lambda c: f"{c}可爱图片",
    lambda c: f"{c}酷炫图片",
    lambda c: f"一张{c}的酷炫的照片",
    lambda c: f"一张{c}的酷炫图片",
    lambda c: f"这是{c}",
    lambda c: f"{c}的好看照片",
    lambda c: f"一张{c}的好看的图片",
    lambda c: f"{c}的好看图片",
    lambda c: f"{c}的照片。",
    lambda c: f"质量差的{c}的照片。",
    lambda c: f"许多{c}的照片。",
    lambda c: f"{c}的雕塑。",
    lambda c: f"难以看到{c}的照片。",
    lambda c: f"{c}的低分辨率照片。",
    lambda c: f"{c}的渲染。",
    lambda c: f"涂鸦{c}。",
    lambda c: f"{c}的糟糕照片。",
    lambda c: f"{c}的裁剪照片。",
    lambda c: f"{c}的纹身。",
    lambda c: f"{c}的刺绣照片。",
    lambda c: f"很难看到{c}的照片。",
    lambda c: f"{c}的明亮照片。",
    lambda c: f"一张干净的{c}的照片。",
    lambda c: f"一张包含{c}的照片。",
    lambda c: f"{c}的深色照片。",
    lambda c: f"{c}的手绘画。",
    lambda c: f"我的{c}的照片。",
    lambda c: f"不自然的{c}的照片。",
    lambda c: f"一张酷的{c}的照片。",
    lambda c: f"{c}的特写照片。",
    lambda c: f"{c}的黑白照片。",
    lambda c: f"一幅{c}的画。",
    lambda c: f"一幅{c}的绘画。",
    lambda c: f"一张{c}的像素照片。",
    lambda c: f"{c}的雕像。",
    lambda c: f"一张{c}的明亮照片。",
    lambda c: f"{c}的裁剪照片。",
    lambda c: f"人造的{c}的照片。",
    lambda c: f"一张关于{c}的照片。",
    lambda c: f"损坏的{c}的jpeg照片。",
    lambda c: f"{c}的模糊照片。",
    lambda c: f"{c}的相片。",
    lambda c: f"一张{c}的好照片。",
    lambda c: f"{c}的渲染照。",
    lambda c: f"视频游戏中的{c}。",
    lambda c: f"一张{c}的照片。",
    lambda c: f"{c}的涂鸦。",
    lambda c: f"{c}的近距离照片。",
    lambda c: f"{c}的折纸。",
    lambda c: f"{c}在视频游戏中。",
    lambda c: f"{c}的草图。",
    lambda c: f"{c}的涂鸦照。",
    lambda c: f"{c}的折纸形状。",
    lambda c: f"低分辨率的{c}的照片。",
    lambda c: f"玩具{c}。",
    lambda c: f"{c}的副本。",
    lambda c: f"{c}的干净的照片。",
    lambda c: f"一张大{c}的照片。",
    lambda c: f"{c}的重现。",
    lambda c: f"一张漂亮的{c}的照片。",
    lambda c: f"一张奇怪的{c}的照片。",
    lambda c: f"模糊的{c}的照片。",
    lambda c: f"卡通{c}。",
    lambda c: f"{c}的艺术作品。",
    lambda c: f"{c}的素描。",
    lambda c: f"刺绣{c}。",
    lambda c: f"{c}的像素照。",
    lambda c: f"{c}的拍照。",
    lambda c: f"{c}的损坏的照片。",
    lambda c: f"高质量的{c}的照片。",
    lambda c: f"毛绒玩具{c}。",
    lambda c: f"漂亮的{c}的照片。",
    lambda c: f"小{c}的照片。",
    lambda c: f"照片是奇怪的{c}。",
    lambda c: f"漫画{c}。",
    lambda c: f"{c}的艺术照。",
    lambda c: f"{c}的图形。",
    lambda c: f"大{c}的照片。",
    lambda c: f"黑白的{c}的照片。",
    lambda c: f"{c}毛绒玩具。",
    lambda c: f"一张{c}的深色照片。",
    lambda c: f"{c}的摄影图。",
    lambda c: f"{c}的涂鸦照。",
    lambda c: f"玩具形状的{c}。",
    lambda c: f"拍了{c}的照片。",
    lambda c: f"酷酷的{c}的照片。",
    lambda c: f"照片里的小{c}。",
    lambda c: f"{c}的刺青。",
    lambda c: f"{c}的可爱的照片。",
    lambda c: f"一张{c}可爱的照片。",
    lambda c: f"{c}可爱图片。",
    lambda c: f"{c}酷炫图片。",
    lambda c: f"一张{c}的酷炫的照片。",
    lambda c: f"一张{c}的酷炫图片。",
    lambda c: f"这是{c}。",
    lambda c: f"{c}的好看照片。",
    lambda c: f"一张{c}的好看的图片。",
    lambda c: f"{c}的好看图片。",
    lambda c: f"一种叫{c}的花的照片",
    lambda c: f"一种叫{c}的食物的照片",
    lambda c: f"{c}的卫星照片"
]

normal_templates = [lambda c: f"{c}的图片"]

flower_templates = [
    lambda c: f"一种叫{c}的花的照片",
    lambda c: f"一种叫{c}的花卉的照片",
    lambda c: f"一种叫{c}的花朵的照片",
    lambda c: f"一种叫{c}的鲜花的照片",
    lambda c: f"一种叫{c}的花的高清图",
    lambda c: f"一种叫{c}的花卉的高清图",
    lambda c: f"一种叫{c}的花朵的高清图",
    lambda c: f"一种叫{c}的鲜花的高清图",
    lambda c: f"一种叫{c}的花的模糊图片",
    lambda c: f"一种叫{c}的花朵的模糊图片",
    lambda c: f"一种叫{c}的花卉的模糊图片",
    lambda c: f"一种叫{c}的鲜花的模糊图片",
    lambda c: f"一种叫{c}的花的缩放图片",
    lambda c: f"一种叫{c}的花朵的缩放图片",
    lambda c: f"一种叫{c}的花卉的缩放图片",
    lambda c: f"一种叫{c}的鲜花的缩放图片",
    lambda c: f"一种叫{c}的花的摄影图",
    lambda c: f"一种叫{c}的花卉的摄影图",
    lambda c: f"一种叫{c}的花朵的摄影图",
    lambda c: f"一种叫{c}的鲜花的摄影图",
    lambda c: f"一种叫{c}的花的近距离照片",
    lambda c: f"一种叫{c}的花朵的近距离照片",
    lambda c: f"一种叫{c}的花卉的近距离照片",
    lambda c: f"一种叫{c}的鲜花的近距离照片",
    lambda c: f"一种叫{c}的花的裁剪照片",
    lambda c: f"一种叫{c}的花朵的裁剪照片",
    lambda c: f"一种叫{c}的花卉的裁剪照片",
    lambda c: f"一种叫{c}的鲜花的裁剪照片",
    lambda c: f"一种叫{c}的花的好看的图片",
    lambda c: f"一种叫{c}的花朵的好看的图片",
    lambda c: f"一种叫{c}的花卉的好看的图片",
    lambda c: f"一种叫{c}的鲜花的好看的图片",
]

food_templates = [
    lambda c: f"一种叫{c}的食物的照片",
    lambda c: f"一种叫{c}的美食的照片",
    lambda c: f"一种叫{c}的菜的照片",
    lambda c: f"一种叫{c}的食物的高清图",
    lambda c: f"一种叫{c}的美食的高清图",
    lambda c: f"一种叫{c}的菜的高清图",
    lambda c: f"一种叫{c}的食物的模糊图片",
    lambda c: f"一种叫{c}的美食的模糊图片",
    lambda c: f"一种叫{c}的菜的模糊图片",
    lambda c: f"一种叫{c}的食物的缩放图片",
    lambda c: f"一种叫{c}的美食的缩放图片",
    lambda c: f"一种叫{c}的菜的缩放图片",
    lambda c: f"一种叫{c}的食物的摄影图",
    lambda c: f"一种叫{c}的美食的摄影图",
    lambda c: f"一种叫{c}的菜的摄影图",
    lambda c: f"一种叫{c}的食物的近距离照片",
    lambda c: f"一种叫{c}的美食的近距离照片",
    lambda c: f"一种叫{c}的菜的近距离照片",
    lambda c: f"一种叫{c}的食物的裁剪照片",
    lambda c: f"一种叫{c}的美食的裁剪照片",
    lambda c: f"一种叫{c}的菜的裁剪照片",
]

aircraft_templates = [
    lambda c: f"{c}，飞机的照片",
    lambda c: f"{c}，飞机的高清图",
    lambda c: f"{c}，飞机的模糊图片",
    lambda c: f"{c}，飞机的缩放图片",
    lambda c: f"{c}，飞机的摄影图",
    lambda c: f"{c}，战斗机的照片",
    lambda c: f"{c}，战斗机的高清图",
    lambda c: f"{c}，战斗机的模糊图片",
    lambda c: f"{c}，战斗机的缩放图片",
    lambda c: f"{c}，战斗机的摄影图",
    lambda c: f"{c}，老飞机的照片",
    lambda c: f"{c}，老飞机的高清图",
    lambda c: f"{c}，老飞机的模糊图片",
    lambda c: f"{c}，老飞机的缩放图片",
    lambda c: f"{c}，老飞机的摄影图",
    lambda c: f"{c}，大飞机的照片",
    lambda c: f"{c}，大飞机的高清图",
    lambda c: f"{c}，大飞机的模糊图片",
    lambda c: f"{c}，大飞机的缩放图片",
    lambda c: f"{c}，大飞机的摄影图",
    lambda c: f"{c}，小飞机的照片",
    lambda c: f"{c}，小飞机的高清图",
    lambda c: f"{c}，小飞机的模糊图片",
    lambda c: f"{c}，小飞机的缩放图片",
    lambda c: f"{c}，小飞机的摄影图",
    lambda c: f"{c}，军用飞机的照片",
    lambda c: f"{c}，军用飞机的高清图",
    lambda c: f"{c}，军用飞机的模糊图片",
    lambda c: f"{c}，军用飞机的缩放图片",
    lambda c: f"{c}，军用飞机的摄影图",
    lambda c: f"{c}，运输机的照片",
    lambda c: f"{c}，运输机的高清图",
    lambda c: f"{c}，运输机的模糊图片",
    lambda c: f"{c}，运输机的缩放图片",
    lambda c: f"{c}，运输机的摄影图",
    lambda c: f"{c}，公务机的照片",
    lambda c: f"{c}，公务机的高清图",
    lambda c: f"{c}，公务机的模糊图片",
    lambda c: f"{c}，公务机的缩放图片",
    lambda c: f"{c}，公务机的摄影图",
    lambda c: f"{c}，客机的照片",
    lambda c: f"{c}，客机的高清图",
    lambda c: f"{c}，客机的模糊图片",
    lambda c: f"{c}，客机的缩放图片",
    lambda c: f"{c}，客机的摄影图",
    lambda c: f"{c}，喷气机的照片",
    lambda c: f"{c}，喷气机的高清图",
    lambda c: f"{c}，喷气机的模糊图片",
    lambda c: f"{c}，喷气机的缩放图片",
    lambda c: f"{c}，喷气机的摄影图",
    lambda c: f"一种叫{c}的飞机的照片",
    lambda c: f"一种叫{c}的飞机的高清图",
    lambda c: f"一种叫{c}的飞机的模糊图片",
    lambda c: f"一种叫{c}的飞机的缩放图片",
    lambda c: f"一种叫{c}的飞机的摄影图",
    lambda c: f"一种叫{c}的战斗机的照片",
    lambda c: f"一种叫{c}的战斗机的高清图",
    lambda c: f"一种叫{c}的战斗机的模糊图片",
    lambda c: f"一种叫{c}的战斗机的缩放图片",
    lambda c: f"一种叫{c}的战斗机的摄影图",
    lambda c: f"一种叫{c}的老飞机的照片",
    lambda c: f"一种叫{c}的老飞机的高清图",
    lambda c: f"一种叫{c}的老飞机的模糊图片",
    lambda c: f"一种叫{c}的老飞机的缩放图片",
    lambda c: f"一种叫{c}的老飞机的摄影图",
    lambda c: f"一种叫{c}的大飞机的照片",
    lambda c: f"一种叫{c}的大飞机的高清图",
    lambda c: f"一种叫{c}的大飞机的模糊图片",
    lambda c: f"一种叫{c}的大飞机的缩放图片",
    lambda c: f"一种叫{c}的大飞机的摄影图",
    lambda c: f"一种叫{c}的小飞机的照片",
    lambda c: f"一种叫{c}的小飞机的高清图",
    lambda c: f"一种叫{c}的小飞机的模糊图片",
    lambda c: f"一种叫{c}的小飞机的缩放图片",
    lambda c: f"一种叫{c}的小飞机的摄影图",
    lambda c: f"一种叫{c}的军用飞机的照片",
    lambda c: f"一种叫{c}的军用飞机的高清图",
    lambda c: f"一种叫{c}的军用飞机的模糊图片",
    lambda c: f"一种叫{c}的军用飞机的缩放图片",
    lambda c: f"一种叫{c}的军用飞机的摄影图",
    lambda c: f"一种叫{c}的运输机的照片",
    lambda c: f"一种叫{c}的运输机的高清图",
    lambda c: f"一种叫{c}的运输机的模糊图片",
    lambda c: f"一种叫{c}的运输机的缩放图片",
    lambda c: f"一种叫{c}的运输机的摄影图",
    lambda c: f"一种叫{c}的公务机的照片",
    lambda c: f"一种叫{c}的公务机的高清图",
    lambda c: f"一种叫{c}的公务机的模糊图片",
    lambda c: f"一种叫{c}的公务机的缩放图片",
    lambda c: f"一种叫{c}的公务机的摄影图",
    lambda c: f"一种叫{c}的客机的照片",
    lambda c: f"一种叫{c}的客机的高清图",
    lambda c: f"一种叫{c}的客机的模糊图片",
    lambda c: f"一种叫{c}的客机的缩放图片",
    lambda c: f"一种叫{c}的客机的摄影图",
    lambda c: f"一种叫{c}的喷气机的照片",
    lambda c: f"一种叫{c}的喷气机的高清图",
    lambda c: f"一种叫{c}的喷气机的模糊图片",
    lambda c: f"一种叫{c}的喷气机的缩放图片",
    lambda c: f"一种叫{c}的喷气机的摄影图",
]

eurosat_templates = [
    lambda c: f"一张{c}的卫星照片",
    lambda c: f"{c}的卫星照片",
    lambda c: f"一张{c}的高清卫星照片",
    lambda c: f"{c}的高清卫星照片",
    lambda c: f"一张{c}的清晰的卫星照片",
    lambda c: f"{c}的清晰的卫星照片",
    lambda c: f"一张{c}的高质量的卫星照片",
    lambda c: f"{c}的高质量的卫星照片",
    lambda c: f"一张{c}的卫星图",
    lambda c: f"{c}的卫星图",
    lambda c: f"一张{c}的高清卫星图",
    lambda c: f"{c}的高清卫星图",
    lambda c: f"一张{c}的清晰的卫星图",
    lambda c: f"{c}的清晰的卫星图",
    lambda c: f"一张{c}的高质量的卫星图",
    lambda c: f"{c}的高质量的卫星图",
    lambda c: f"一张{c}的卫星图片",
    lambda c: f"{c}的卫星图片",
    lambda c: f"一张{c}的高清卫星图片",
    lambda c: f"{c}的高清卫星图片",
    lambda c: f"一张{c}的清晰的卫星图片",
    lambda c: f"{c}的清晰的卫星图片",
    lambda c: f"一张{c}的高质量的卫星图片",
    lambda c: f"{c}的高质量的卫星图片",
]

hatefulmemes_templates = [
    lambda c: f"一个{c}",
    lambda c: f"{c}",
]

kitti_templates = [
    lambda c: f"照片里{c}",
    lambda c: f"图片里{c}",
    lambda c: f"{c}",
]

cars_templates = [
    lambda c: f"一张{c}的照片",
    lambda c: f"一张我的{c}的照片",
    lambda c: f"我爱我的{c}",
    lambda c: f"一张我肮脏的{c}的照片",
    lambda c: f"一张我干净的{c}的照片",
    lambda c: f"一张我新买的{c}的照片",
    lambda c: f"一张我旧的{c}的照片",
]

dtd_templates = [
    lambda c: f"一张{c}纹理的照片",
    lambda c: f"一张{c}图案的照片",
    lambda c: f"一张{c}物体的照片",
    lambda c: f"一张{c}纹理的图片",
    lambda c: f"一张{c}图案的图片",
    lambda c: f"一张{c}物体的图片",
]

country211_templates = [
    lambda c: f"一张在{c}拍的照片",
    lambda c: f"一张在{c}旅行时拍的照片",
    lambda c: f"一张我家乡{c}的照片",
    lambda c: f"一张展示{c}风光的照片",
]

patch_templates = [
    lambda c: f"一张{c}的医疗照片",
    lambda c: f"一张{c}的ct照片",
    lambda c: f"一张{c}的化验照片",
]

pet_templates = [
    lambda c: f"一种叫{c}的宠物的照片",
    lambda c: f"一种叫{c}的宠物的图片",
    lambda c: f"一种叫{c}的宠物的可爱图片",
    lambda c: f"一种叫{c}的宠物的高清图片",
    lambda c: f"一种叫{c}的宠物的模糊图片",
    lambda c: f"一种叫{c}的宠物的特写照片",
]

cifar100_templates = [
    lambda c: f"一张{c}的照片",
    lambda c: f"一张{c}的模糊照片",
    lambda c: f"一张{c}",
    lambda c: f"一张{c}的低对比度照片",
    lambda c: f"一张{c}的高对比度照片",
    lambda c: f"一张{c}的好照片",
    lambda c: f"一张小{c}的照片",
    lambda c: f"一张大{c}的照片",
    lambda c: f"一张{c}的黑白照片",
    lambda c: f"一张{c}的低对比度的照片",
    lambda c: f"一张{c}的高对比度的照片",
]

caltech101_templates = [
    lambda c: f"{c}的照片",
    lambda c: f"{c}的绘画",
    lambda c: f"{c}的塑料",
    lambda c: f"{c}的雕像",
    lambda c: f"{c}的草图",
    lambda c: f"{c}的刺青",
    lambda c: f"{c}的玩具",
    lambda c: f"{c}的演绎",
    lambda c: f"{c}的装饰",
    lambda c: f"{c}的卡通画",
    lambda c: f"{c}在游戏中",
    lambda c: f"一个豪华的{c}.",
    lambda c: f"{c}的折纸",
    lambda c: f"{c}的艺术画",
    lambda c: f"{c}的涂鸦画",
    lambda c: f"{c}的画",
]

fer_templates = [
    lambda c: f"一张表情{c}的照片",
    lambda c: f"一张表达{c}情绪的照片",
    lambda c: f"一张看起来很{c}的照片",
    lambda c: f"他的脸看起来{c}",
    lambda c: f"他们看起来很{c}",
]


# -*- coding: utf-8 -*-
"""
This script runs to compare the inference speed between Pytorch (CPU/GPU), ONNX (CPU/GPU) and TensorRT models.
"""

import argparse
import torch
from PIL import Image
import cn_clip.clip as clip
from cn_clip.clip.utils import create_model, _MODEL_INFO, image_transform
from cn_clip.training.main import convert_models_to_fp32, convert_weights
from cn_clip.deploy.benchmark_utils import track_infer_time, print_timings


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-arch", 
        required=True, 
        choices=["ViT-B-16", "ViT-L-14", "ViT-L-14-336", "ViT-H-14", "RN50"],
        help="Specify the architecture (model scale) of Chinese-CLIP model for speed comparison."
    )
    parser.add_argument('--pytorch-ckpt', type=str, default=None, 
                        help='The file path of pytorch checkpoint, if not set, the program will not run Pytorch model.')
    parser.add_argument('--pytorch-precision', choices=["fp16", "fp32"], default="fp16", 
                        help='Flag for Pytorch model float point precision, default to using FP16 for evaluation.')
    
    parser.add_argument('--onnx-image-model', type=str, default=None,
                        help='The file path of onnx image model, if not set, the program will not run image ONNX model.')
    parser.add_argument('--onnx-text-model', type=str, default=None,
                        help='The file path of onnx text model, if not set, the program will not run text ONNX model.')

    parser.add_argument('--tensorrt-image-model', type=str, default=None,
                        help='The file path of image TensorRT model, if not set, the program will not run image TensorRT model.')
    parser.add_argument('--tensorrt-text-model', type=str, default=None,
                        help='The file path of text TensorRT model, if not set, the program will not run text TensorRT model.')

    parser.add_argument('--batch-size', default=1, type=int, help='The batch-size of the inference input. Default to 1.')
    parser.add_argument('--n', default=100, type=int, help='The iteration number for inference speed test. Default to 100.')
    parser.add_argument('--warmup', default=10, type=int, help='Warmup iterations. Default to 10.')

    parser.add_argument("--device", choices=["cuda", "cpu"], default="cuda", help="CPU or GPU speed test. Default to cuda",)
    parser.add_argument(
        "--context-length", type=int, default=52, help="The padded length of input text (include [CLS] & [SEP] tokens). Default to 52."
    )

    args = parser.parse_args()

    return args


def prepare_pytorch_model(args):
    with open(args.pytorch_ckpt, 'rb') as opened_file:
        checkpoint = torch.load(opened_file, map_location="cpu")
    pt_model = create_model(_MODEL_INFO[args.model_arch]['struct'], checkpoint)
    pt_model.eval()
    if args.pytorch_precision == "fp16":
        convert_weights(pt_model)
    else:
        convert_models_to_fp32(pt_model)
    if args.device == "cuda":
        pt_model.cuda()
    return pt_model


onnx_execution_provider_map = {
    "cpu": "CPUExecutionProvider",
    "cuda": "CUDAExecutionProvider",
}


if __name__ == '__main__':
    args = parse_args()
    
    # Log params.
    print("Params:")
    for name in sorted(vars(args)):
        val = getattr(args, name)
        print(f"  {name}: {val}")

    preprocess = image_transform(_MODEL_INFO[args.model_arch]['input_resolution'])
    image = torch.vstack([preprocess(Image.open("examples/pokemon.jpeg")).unsqueeze(0)] * args.batch_size)
    text = torch.vstack([clip.tokenize(["杰尼龟"], context_length=args.context_length)] * args.batch_size)
    if args.device == "cuda":
        image = image.cuda()
        text = text.cuda()

    # test the image feature extraction
    print("Begin the image feature extraction speed test...")
    if args.pytorch_ckpt:
        print(f"Prepare the Pytorch model from {args.pytorch_ckpt}")
        pt_model = prepare_pytorch_model(args)
        for i in range(args.warmup):
            pytorch_output = pt_model(image=image, text=None)
        print("Forward the Pytorch image model...")
        time_buffer = list()
        for i in range(args.n):
            with track_infer_time(time_buffer):
                pytorch_output = pt_model(image=image, text=None)
        print_timings(name=f"Pytorch image inference speed (batch-size: {args.batch_size}):", timings=time_buffer)
        del pt_model

    if args.onnx_image_model:
        import onnxruntime
        print(f'Prepare the ONNX image model from {args.onnx_image_model}')
        sess_options = onnxruntime.SessionOptions()
        run_options = onnxruntime.RunOptions()
        run_options.log_severity_level = 2
        session = onnxruntime.InferenceSession(args.onnx_image_model,
                                                sess_options=sess_options,
                                                providers=[onnx_execution_provider_map[args.device]]
                                                )
        for i in range(args.warmup):
            onnx_output = session.run(["unnorm_image_features"], {"image": image.cpu().numpy()})
        print("Forward the ONNX image model...")
        time_buffer = list()
        for i in range(args.n):
            with track_infer_time(time_buffer):
                onnx_output = session.run(["unnorm_image_features"], {"image": image.cpu().numpy()})
        print_timings(name=f"ONNX image inference speed (batch-size: {args.batch_size}):", timings=time_buffer)
        del session

    if args.tensorrt_image_model and args.device == "cuda":
        from cn_clip.deploy.tensorrt_utils import TensorRTModel
        print(f'Prepare the TensorRT image model from {args.tensorrt_image_model}')
        trt_model = TensorRTModel(args.tensorrt_image_model)
        for i in range(args.warmup):
            trt_output = trt_model(inputs={"image": image})
        time_buffer = list()
        print("Forward the TensorRT image model...")
        for i in range(args.n):
            with track_infer_time(time_buffer):
                trt_output = trt_model(inputs={"image": image})
        print_timings(name=f"TensorRT image inference speed (batch-size: {args.batch_size}):", timings=time_buffer)
        del trt_model


    # test the image feature extraction
    print("Begin the text feature extraction speed test...")
    if args.pytorch_ckpt:
        print(f"Prepare the Pytorch model from {args.pytorch_ckpt}")
        pt_model = prepare_pytorch_model(args)

        for i in range(args.warmup):
            pytorch_output = pt_model(image=None, text=text)
        print("Forward the Pytorch text model...")
        time_buffer = list()
        for i in range(args.n):
            with track_infer_time(time_buffer):
                pytorch_output = pt_model(image=None, text=text)
        print_timings(name=f"Pytorch text inference speed (batch-size: {args.batch_size}):", timings=time_buffer)
        del pt_model

    if args.onnx_text_model:
        import onnxruntime
        print(f'Prepare the ONNX text model from {args.onnx_text_model}')
        sess_options = onnxruntime.SessionOptions()
        run_options = onnxruntime.RunOptions()
        run_options.log_severity_level = 2
        session = onnxruntime.InferenceSession(args.onnx_text_model,
                                                sess_options=sess_options,
                                                providers=[onnx_execution_provider_map[args.device]]
                                                )
        for i in range(args.warmup):
            onnx_output = session.run(["unnorm_text_features"], {"text": text.cpu().numpy()})
        print("Forward the ONNX text model...")
        time_buffer = list()
        for i in range(args.n):
            with track_infer_time(time_buffer):
                onnx_output = session.run(["unnorm_text_features"], {"text": text.cpu().numpy()})
        print_timings(name=f"ONNX text inference speed (batch-size: {args.batch_size}):", timings=time_buffer)
        del session

    if args.tensorrt_text_model and args.device == "cuda":
        from cn_clip.deploy.tensorrt_utils import TensorRTModel
        print(f'Prepare the TensorRT text model from {args.tensorrt_text_model}')        
        trt_model = TensorRTModel(args.tensorrt_text_model)
        for i in range(args.warmup):
            trt_output = trt_model(inputs={'text': text})
        print("Forward the TensorRT text model...")
        time_buffer = list()
        for i in range(args.n):
            with track_infer_time(time_buffer):
                trt_output = trt_model(inputs={'text': text})
        print_timings(name=f"TensorRT text inference speed (batch-size: {args.batch_size}):", timings=time_buffer)
        del trt_model
    
    print("Done!")